# Copyright 2024-2025 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import threading
import json

from zuul.zk import ZooKeeperClient
from zuul.zk.cache import ZuulTreeCache
from zuul.zk.components import (
    ComponentRegistry,
    COMPONENT_REGISTRY
)
from tests.base import (
    BaseTestCase,
    iterate_timeout,
    ZOOKEEPER_SESSION_TIMEOUT,
)

from kazoo.protocol.states import KazooState


class SimpleTreeCacheObject:
    def __init__(self, root, key, data, zstat):
        self.key = key
        self.data = json.loads(data)
        self._zstat = zstat
        self.path = '/'.join((root.rstrip("/"), *key))
        self.children = {}

    def _updateFromRaw(self, context, data, zstat, extra):
        self.data = json.loads(data)
        self._zstat = zstat


class SimpleTreeCache(ZuulTreeCache):
    def objectFromRaw(self, key, data, zstat):
        return SimpleTreeCacheObject(self.root, key, data, zstat)

    def updateFromRaw(self, obj, key, data, zstat):
        obj._updateFromRaw(None, data, zstat, None)

    def parsePath(self, path):
        object_path = path[len(self.root):].strip("/")
        parts = object_path.split('/')
        if not parts:
            return None, False
        return tuple(parts), True


class SimpleSubnodeTreeCache(SimpleTreeCache):
    def preCacheHook(self, event, exists, data=None, stat=None):
        parts, shouldFetch = self.parsePath(event.path)
        if len(parts) > 1:
            cache_key = (parts[0],)
            if exists:
                self._cached_objects[cache_key].children[parts] = data
            else:
                self._cached_objects[cache_key].children.pop(parts)
            return self.STOP_OBJECT_UPDATE


class TestTreeCache(BaseTestCase):
    # A very simple smoke test of the tree cache

    def setUp(self):
        super().setUp()

        self.setupZK()

        self.zk_client = ZooKeeperClient(
            self.zk_chroot_fixture.zk_hosts,
            tls_cert=self.zk_chroot_fixture.zookeeper_cert,
            tls_key=self.zk_chroot_fixture.zookeeper_key,
            tls_ca=self.zk_chroot_fixture.zookeeper_ca,
            timeout=ZOOKEEPER_SESSION_TIMEOUT,
        )
        self.addCleanup(self.zk_client.disconnect)
        self.zk_client.connect()
        self.setupModelPin()
        self.component_registry = ComponentRegistry(self.zk_client)
        # We don't have any other component to initialize the global
        # registry in these tests, so we do it ourselves.
        COMPONENT_REGISTRY.create(self.zk_client)

    def waitForCache(self, cache, contents):
        paths = set(contents.keys())
        for _ in iterate_timeout(10, 'cache to sync'):
            cached_paths = cache._cached_paths.copy()
            cached_paths.discard(cache.root)
            object_paths = set(
                [x.path for x in cache._cached_objects.values()])
            if paths == cached_paths == object_paths:
                found = True
                for obj in cache._cached_objects.values():
                    if contents[obj.path] != obj.data:
                        found = False
                if found:
                    return

    def _test_tree_cache(self, async_worker):
        client = self.zk_client.client
        data = b'{}'
        client.create('/test', data)
        client.create('/test/foo', data)
        cache = SimpleTreeCache(self.zk_client, "/test",
                                async_worker=async_worker)
        self.waitForCache(cache, {
            '/test/foo': {},
        })
        client.create('/test/bar', data)
        self.waitForCache(cache, {
            '/test/foo': {},
            '/test/bar': {},
        })
        client.set('/test/bar', b'{"value":1}')
        self.waitForCache(cache, {
            '/test/foo': {},
            '/test/bar': {'value': 1},
        })
        client.delete('/test/bar')
        self.waitForCache(cache, {
            '/test/foo': {},
        })

        # Simulate a change happening while the state was lost
        cache._cached_paths.add('/test/bar')
        cache._sessionListener(KazooState.LOST)
        cache._sessionListener(KazooState.CONNECTED)
        self.waitForCache(cache, {
            '/test/foo': {},
        })

        # Simulate a change happening while the state was suspended
        cache._cached_paths.add('/test/bar')
        cache._sessionListener(KazooState.SUSPENDED)
        cache._sessionListener(KazooState.CONNECTED)
        self.waitForCache(cache, {
            '/test/foo': {},
        })

    def test_tree_cache_async(self):
        self._test_tree_cache(async_worker=True)

    def test_tree_cache_sync(self):
        self._test_tree_cache(async_worker=False)

    def test_tree_cache_root(self):
        client = self.zk_client.client
        data = b'{}'
        client.create('/foo', data)
        cache = SimpleTreeCache(self.zk_client, "/")
        for _ in iterate_timeout(10, 'cache to sync'):
            cached_paths = cache._cached_paths.copy()
            cached_paths.discard(cache.root)
            object_paths = set(
                [x.path for x in cache._cached_objects.values()])
            if ('/foo' in cached_paths and
                '/foo' in object_paths):
                break

    def test_tree_cache_subnode(self):
        client = self.zk_client.client
        data = b'{}'
        client.create('/test', data)
        client.create('/test/foo', data)
        cache = SimpleSubnodeTreeCache(self.zk_client, "/test")
        self.waitForCache(cache, {
            '/test/foo': {},
        })
        foo = cache._cached_objects[('foo',)]

        # Subnode that is processed and cached as part of foo
        oof_data = b'{"value":1}'
        oof_key = ('foo', 'oof')
        client.create('/test/foo/oof', oof_data)
        for _ in iterate_timeout(10, 'cache to sync'):
            if foo.children.get(oof_key) == oof_data:
                break

        self.assertEqual(cache._cached_paths, {'/test/foo', '/test/foo/oof'})

        # Simulate a change happening while the state was suspended
        foo.children[oof_key] = b"outdated"
        cache._sessionListener(KazooState.SUSPENDED)
        cache._sessionListener(KazooState.CONNECTED)
        for _ in iterate_timeout(10, 'cache to sync'):
            if foo.children[oof_key] == oof_data:
                break

        # Simulate a change happening while the state was lost
        cache._cached_paths.add('/test/foo/bar')
        bar_key = ('foo', 'bar')
        foo.children[bar_key] = b"deleted"
        cache._sessionListener(KazooState.LOST)
        cache._sessionListener(KazooState.CONNECTED)
        for _ in iterate_timeout(10, 'cache to sync'):
            if bar_key not in foo.children:
                break

        self.assertEqual(cache._cached_paths, {'/test/foo', '/test/foo/oof'})

        # Recursively delete foo and make sure the cache is empty afterwards
        client.delete("/test/foo", recursive=True)
        self.waitForCache(cache, {})
        self.assertEqual(cache._cached_paths, set())
        self.assertEqual(cache._cached_objects, {})

    def test_tree_cache_qsize_warning(self):
        with self.assertLogs('zuul.zk.ZooKeeper', level='DEBUG') as logs:
            cache = SimpleTreeCache(self.zk_client, "/test")
            cache._last_event_warning = 0
            cache._last_playback_warning = 0
            cache.qsize_warning_threshold = -1

            data = b'{}'
            client = self.zk_client.client
            client.create('/test/foo', data)
            self.waitForCache(cache, {
                '/test/foo': {},
            })

            found_event_warning = False
            found_playback_warning = False
            for line in logs.output:
                self.log.debug("Received %s", str(line))
                if 'Event queue size for cache' in str(line):
                    found_event_warning = True
                if 'Playback queue size for cache' in str(line):
                    found_playback_warning = True
            self.assertTrue(found_event_warning)
            self.assertTrue(found_playback_warning)

    def test_tree_cache_max_zxid(self):
        client = self.zk_client.client
        data = b'{}'
        client.create('/test', b'')
        client.create('/test/bar', data)

        cache = SimpleTreeCache(self.zk_client, "/test")
        self.waitForCache(cache, {
            '/test/bar': {},
        })

        orig_zxid = cache.max_zxid
        self.assertGreater(cache.max_zxid, -1)
        client.create('/test/foo', data)
        self.waitForCache(cache, {
            '/test/bar': {},
            '/test/foo': {},
        })
        self.assertGreater(cache.max_zxid, orig_zxid)

    def test_wait_for_sync(self):
        # Test that waitForSync works during a crash
        client = self.zk_client.client
        data = b'{}'
        client.create('/test', data)
        client.create('/test/foo', data)
        cache = SimpleTreeCache(self.zk_client, "/test",
                                async_worker=True)
        self.waitForCache(cache, {
            '/test/foo': {},
        })

        waiting_event = threading.Event()
        go_event = threading.Event()
        orig_wait = cache.waitForSync

        def wait(*args, **kw):
            waiting_event.set()
            go_event.wait()
            return orig_wait(*args, **kw)
        self.patch(cache, 'waitForSync', wait)

        # Prevent waitForSync from running
        go_event.clear()

        def waiter():
            self.log.debug("Wait for sync")
            cache.waitForSync()
            self.log.debug("Done wait for sync")

        waiter_thread = threading.Thread(target=waiter)
        waiter_thread.start()
        waiting_event.wait()
        client.create('/test/bar', data)
        cache._sessionListener(KazooState.LOST)
        cache._sessionListener(KazooState.CONNECTED)

        # Allow waitForSync to run
        go_event.set()
        waiter_thread.join()
        cache.ensureReady()
        self.waitForCache(cache, {
            '/test/foo': {},
            '/test/bar': {},
        })



# This file is derived from the Kazoo project
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from collections import namedtuple

from kazoo.protocol.serialization import (
    int_struct,
    write_string,
)


class RemoveWatches(namedtuple("RemoveWatches", "path watcher_type")):
    type = 18

    def serialize(self):
        b = bytearray()
        b.extend(write_string(self.path))
        b.extend(int_struct.pack(self.watcher_type))
        return b

    @classmethod
    def deserialize(cls, bytes, offset):
        return None


class AddWatch(namedtuple("AddWatch", "path watcher mode")):
    type = 106

    def serialize(self):
        b = bytearray()
        b.extend(write_string(self.path))
        b.extend(int_struct.pack(self.mode))
        return b

    @classmethod
    def deserialize(cls, bytes, offset):
        return None



# Copyright 2021 BMW Group
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging

from zuul.model import MergeRequest
from zuul.zk.job_request_queue import JobRequestQueue


class MergerApi(JobRequestQueue):
    log = logging.getLogger("zuul.MergerApi")
    request_class = MergeRequest

    def __init__(self, client, use_cache=True, merge_request_callback=None):
        root = '/zuul/merger'
        super().__init__(client, root, use_cache, merge_request_callback)



# Copyright 2025 Acme Gating, LLC
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.

from zuul.ansible import paths, stream_setup

normal = paths._import_ansible_action_plugin("normal")

# Map the various names to our custom module
module_map = {
    'win_shell': 'zuul_win_shell',
    'win_command': 'zuul_win_command',
    'ansible.legacy.win_shell': 'zuul_win_shell',
    'ansible.legacy.win_command': 'zuul_win_command',
    'ansible.windows.win_shell': 'zuul_win_shell',
    'ansible.windows.win_command': 'zuul_win_command',
}

# Never let users specify these directly
module_blocklist = [
    'zuul_win_shell',
    'zuul_win_command',
]


class ActionModule(normal.ActionModule):

    def run(self, tmp=None, task_vars=None):
        module_name = self._task.action
        if module_name in module_blocklist:
            raise Exception(f"Module {module_name} may not be used directly")
        if module_name in module_map:
            if not stream_setup.zuul_console_disabled(self):
                stream_setup.stream_setup_run(self, task_vars)
        return super(ActionModule, self).run(tmp, task_vars)

    def _execute_module(self, module_name=None, *args, **kw):
        if module_name is None:
            module_name = self._task.action
        # If the user has not disabled the zuul console, then switch
        # to our custom module.
        if module_name in module_map:
            if not stream_setup.zuul_console_disabled(self):
                module_name = module_map.get(module_name)
        return super(ActionModule, self)._execute_module(
            module_name=module_name, *args, **kw)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import subprocess

from tests.base import ZuulTestCase
from zuul.executor.server import SshAgent


class TestSshAgent(ZuulTestCase):
    tenant_config_file = 'config/single-tenant/main.yaml'

    def test_ssh_agent(self):
        # Need a private key to add
        env_copy = dict(os.environ)
        # DISPLAY and SSH_ASKPASS will cause interactive test runners to get a
        # surprise
        if 'DISPLAY' in env_copy:
            del env_copy['DISPLAY']
        if 'SSH_ASKPASS' in env_copy:
            del env_copy['SSH_ASKPASS']

        agent = SshAgent()
        agent.start()
        env_copy.update(agent.env)

        pub_key_file = '{}.pub'.format(self.private_key_file)
        pub_key = None
        with open(pub_key_file) as pub_key_f:
            pub_key = pub_key_f.read().split('== ')[0]

        agent.add(self.private_key_file)
        keys = agent.list()
        self.assertEqual(1, len(keys))
        self.assertEqual(keys[0].split('== ')[0], pub_key)
        agent.remove(self.private_key_file)
        keys = agent.list()
        self.assertEqual([], keys)
        agent.stop()
        # Agent is now dead and thus this should fail
        with open('/dev/null') as devnull:
            self.assertRaises(subprocess.CalledProcessError,
                              subprocess.check_call,
                              ['ssh-add', self.private_key_file],
                              env=env_copy,
                              stderr=devnull)



# Copyright 2015 GoodData
# Copyright (c) 2017 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from tests.base import ZuulTestCase


class TestGerritAndGithub(ZuulTestCase):
    config_file = 'zuul-connections-gerrit-and-github.conf'
    tenant_config_file = 'config/multi-driver/main.yaml'

    # Those tests are also using the fake github implementation which
    # means that every scheduler gets a different fake github instance.
    # Thus, assertions might fail depending on which scheduler did the
    # interaction with Github.
    scheduler_count = 1

    def test_multiple_project_gerrit_and_github(self):
        self.executor_server.hold_jobs_in_build = True

        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        B = self.fake_github.openFakePullRequest('org/project1', 'master', 'B')
        self.fake_github.emitEvent(B.getPullRequestOpenedEvent())
        self.waitUntilSettled()

        self.assertEqual(2, len(self.builds))
        self.assertEqual('project-gerrit', self.builds[0].name)
        self.assertEqual('project1-github', self.builds[1].name)
        self.assertTrue(self.builds[0].hasChanges(A))
        self.assertTrue(self.builds[1].hasChanges(B))

        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()

        # Check on reporting results
        # github should have a success status (only).
        statuses = self.fake_github.getCommitStatuses(
            'org/project1', B.head_sha)
        self.assertEqual(1, len(statuses))
        self.assertEqual('success', statuses[0]['state'])

        # gerrit should have only reported twice, on start and success
        self.assertEqual(A.reported, 2)



# Copyright 2018 BMW Car IT GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging

from zuul.executor.sensors import SensorInterface


class PauseSensor(SensorInterface):
    log = logging.getLogger("zuul.executor.sensor.pause")

    def __init__(self, statsd, base_key, pause=False):
        super().__init__(statsd, base_key)
        self.pause = pause

    def isOk(self):
        if self.statsd:
            self.statsd.gauge(self.base_key + '.pause', int(self.pause))
        if self.pause:
            return False, 'paused'
        else:
            return True, 'running'



# Copyright 2017 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
import urllib

from bs4 import BeautifulSoup
import testtools

from tests.base import ZuulTestCase, WebProxyFixture
from tests.base import ZuulWebFixture


class TestWebURLs(ZuulTestCase):
    tenant_config_file = 'config/single-tenant/main.yaml'

    def setUp(self):
        super(TestWebURLs, self).setUp()
        self.web = self.useFixture(
            ZuulWebFixture(self.config, self.test_config,
                           self.additional_event_queues, self.upstream_root,
                           self.poller_events,
                           self.git_url_with_auth, self.addCleanup,
                           self.test_root))

    def _get(self, port, uri):
        url = "http://localhost:{}{}".format(port, uri)
        self.log.debug("GET {}".format(url))
        req = urllib.request.Request(url)
        try:
            f = urllib.request.urlopen(req)
        except urllib.error.HTTPError:
            raise Exception("Error on URL {}".format(url))
        return f.read()

    def _crawl(self, url):
        page = self._get(self.port, url)
        page = BeautifulSoup(page, 'html.parser')
        for (tag, attr) in [
                ('script', 'src'),
                ('link', 'href'),
                ('a', 'href'),
                ('img', 'src'),
        ]:
            for item in page.find_all(tag):
                suburl = item.get(attr)
                if tag == 'script' and suburl is None:
                    # There can be an embedded script
                    continue
                if suburl.startswith('/'):
                    suburl = suburl[1:]
                link = urllib.parse.urljoin(url, suburl)
                self._get(self.port, link)


class TestDirect(TestWebURLs):
    # Test directly accessing the zuul-web server with no proxy
    def setUp(self):
        super(TestDirect, self).setUp()
        self.port = self.web.port

    def test_status_page(self):
        self._crawl('/')
        self._crawl('/t/tenant-one/status')

    def test_forbidden(self):
        for url in [
                '/?path=/etc/os-release',
                '/?path=../__init__.py',
                '/../__init__.py',
        ]:
            with testtools.ExpectedException(Exception):
                self._get(self.port, url).decode('utf-8')
        index = self._get(self.port, '/zuul/').decode('utf-8')
        for url in [
                '/etc/os-release',
                '//etc/os-release',
                '/%2fetc%2fos-release',
                '/%2e%2e%2f__init__.py',
        ]:
            data = self._get(self.port, url).decode('utf-8')
            self.assertEqual(index, data)


class TestWhiteLabel(TestWebURLs):
    # Test a zuul-web behind a whitelabel proxy (i.e., what
    # zuul.openstack.org does).
    def setUp(self):
        super(TestWhiteLabel, self).setUp()
        rules = [
            ('^/(.*)$', 'http://localhost:{}/\\1'.format(self.web.port)),
        ]
        self.proxy = self.useFixture(WebProxyFixture(rules))
        self.port = self.proxy.port

    def test_status_page(self):
        self._crawl('/')
        self._crawl('/status')

    def test_forbidden(self):
        for url in [
                '/?path=/etc/os-release',
                '/?path=../__init__.py',
        ]:
            with testtools.ExpectedException(Exception):
                self._get(self.port, url).decode('utf-8')
        index = self._get(self.port, '/zuul/').decode('utf-8')
        for url in [
                '/etc/os-release',
                '//etc/os-release',
                '/%2fetc%2fos-release',
                '/../__init__.py',
                '/%2e%2e%2f__init__.py',
        ]:
            data = self._get(self.port, url).decode('utf-8')
            self.assertEqual(index, data)


class TestWhiteLabelAPI(TestWebURLs):
    # Test a zuul-web behind a whitelabel proxy (i.e., what
    # zuul.openstack.org does).
    def setUp(self):
        super(TestWhiteLabelAPI, self).setUp()
        rules = [
            ('^/api/(.*)$',
             'http://localhost:{}/api/tenant/tenant-one/\\1'.format(
                 self.web.port)),
        ]
        self.proxy = self.useFixture(WebProxyFixture(rules))
        self.port = self.proxy.port

    def test_info(self):
        info = json.loads(self._get(self.port, '/api/info').decode('utf-8'))
        self.assertEqual('tenant-one', info['info']['tenant'])


class TestSuburl(TestWebURLs):
    # Test a zuul-web mounted on a suburl (i.e., what software factory
    # does).
    def setUp(self):
        super(TestSuburl, self).setUp()
        rules = [
            ('^/zuul/(.*)$', 'http://localhost:{}/\\1'.format(
                self.web.port)),
        ]
        self.proxy = self.useFixture(WebProxyFixture(rules))
        self.port = self.proxy.port

    def test_status_page(self):
        self._crawl('/zuul/')

    def test_forbidden(self):
        for url in [
                '/zuul/?path=/etc/os-release',
                '/zuul/?path=../__init__.py',
        ]:
            with testtools.ExpectedException(Exception):
                self._get(self.port, url).decode('utf-8')
        index = self._get(self.port, '/zuul/').decode('utf-8')
        for url in [
                '/zuul/etc/os-release',
                '/zuul//etc/os-release',
                '/zuul/%2fetc%2fos-release',
                '/zuul/../__init__.py',
                '/zuul/%2e%2e%2f__init__.py',
        ]:
            data = self._get(self.port, url).decode('utf-8')
            self.assertEqual(index, data)



# Copyright 2012-2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import time

from zuul.model import PromoteEvent

from tests.base import ZuulTestCase, simple_layout, okay_tracebacks


class TestRequirementsApprovalNewerThan(ZuulTestCase):
    """Requirements with a newer-than comment requirement"""

    tenant_config_file = 'config/requirements/newer-than/main.yaml'

    def test_pipeline_require_approval_newer_than(self):
        "Test pipeline requirement: approval newer than"
        return self._test_require_approval_newer_than('org/project1',
                                                      'project1-job')

    def test_trigger_require_approval_newer_than(self):
        "Test trigger requirement: approval newer than"
        return self._test_require_approval_newer_than('org/project2',
                                                      'project2-job')

    def _test_require_approval_newer_than(self, project, job):
        A = self.fake_gerrit.addFakeChange(project, 'master', 'A')
        # A comment event that we will keep submitting to trigger
        comment = A.addApproval('Code-Review', 2, username='nobody')
        self.fake_gerrit.addEvent(comment)
        self.waitUntilSettled()
        # No +1 from Jenkins so should not be enqueued
        self.assertEqual(len(self.history), 0)

        # Add a too-old +1, should not be enqueued
        A.addApproval('Verified', 1, username='jenkins',
                      granted_on=time.time() - 72 * 60 * 60)
        self.fake_gerrit.addEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 0)

        # Add a recent +1
        self.fake_gerrit.addEvent(A.addApproval('Verified', 1,
                                                username='jenkins'))
        self.fake_gerrit.addEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)
        self.assertEqual(self.history[0].name, job)


class TestRequirementsApprovalOlderThan(ZuulTestCase):
    """Requirements with a older-than comment requirement"""

    tenant_config_file = 'config/requirements/older-than/main.yaml'

    def test_pipeline_require_approval_older_than(self):
        "Test pipeline requirement: approval older than"
        return self._test_require_approval_older_than('org/project1',
                                                      'project1-job')

    def test_trigger_require_approval_older_than(self):
        "Test trigger requirement: approval older than"
        return self._test_require_approval_older_than('org/project2',
                                                      'project2-job')

    def _test_require_approval_older_than(self, project, job):
        A = self.fake_gerrit.addFakeChange(project, 'master', 'A')
        # A comment event that we will keep submitting to trigger
        comment = A.addApproval('Code-Review', 2, username='nobody')
        self.fake_gerrit.addEvent(comment)
        self.waitUntilSettled()
        # No +1 from Jenkins so should not be enqueued
        self.assertEqual(len(self.history), 0)

        # Add a recent +1 which should not be enqueued
        A.addApproval('Verified', 1)
        self.fake_gerrit.addEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 0)

        # Add an old +1 which should be enqueued
        A.addApproval('Verified', 1, username='jenkins',
                      granted_on=time.time() - 72 * 60 * 60)
        self.fake_gerrit.addEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)
        self.assertEqual(self.history[0].name, job)


class TestRequirementsUserName(ZuulTestCase):
    """Requirements with a username requirement"""

    tenant_config_file = 'config/requirements/username/main.yaml'

    def test_pipeline_require_approval_username(self):
        "Test pipeline requirement: approval username"
        return self._test_require_approval_username('org/project1',
                                                    'project1-job')

    def test_trigger_require_approval_username(self):
        "Test trigger requirement: approval username"
        return self._test_require_approval_username('org/project2',
                                                    'project2-job')

    def _test_require_approval_username(self, project, job):
        A = self.fake_gerrit.addFakeChange(project, 'master', 'A')
        # A comment event that we will keep submitting to trigger
        comment = A.addApproval('Code-Review', 2, username='nobody')
        self.fake_gerrit.addEvent(comment)
        self.waitUntilSettled()
        # No approval from Jenkins so should not be enqueued
        self.assertEqual(len(self.history), 0)

        # Add an approval from Jenkins
        A.addApproval('Verified', 1, username='jenkins')
        self.fake_gerrit.addEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)
        self.assertEqual(self.history[0].name, job)


class TestRequirementsEmail(ZuulTestCase):
    """Requirements with a email requirement"""

    tenant_config_file = 'config/requirements/email/main.yaml'

    def test_pipeline_require_approval_email(self):
        "Test pipeline requirement: approval email"
        return self._test_require_approval_email('org/project1',
                                                 'project1-job')

    def test_trigger_require_approval_email(self):
        "Test trigger requirement: approval email"
        return self._test_require_approval_email('org/project2',
                                                 'project2-job')

    def _test_require_approval_email(self, project, job):
        A = self.fake_gerrit.addFakeChange(project, 'master', 'A')
        # A comment event that we will keep submitting to trigger
        comment = A.addApproval('Code-Review', 2, username='nobody')
        self.fake_gerrit.addEvent(comment)
        self.waitUntilSettled()
        # No approval from Jenkins so should not be enqueued
        self.assertEqual(len(self.history), 0)

        # Add an approval from Jenkins
        A.addApproval('Verified', 1, username='jenkins')
        self.fake_gerrit.addEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)
        self.assertEqual(self.history[0].name, job)


class TestRequirementsVote1(ZuulTestCase):
    """Requirements with a voting requirement"""

    tenant_config_file = 'config/requirements/vote1/main.yaml'

    def test_pipeline_require_approval_vote1(self):
        "Test pipeline requirement: approval vote with one value"
        return self._test_require_approval_vote1('org/project1',
                                                 'project1-job')

    def test_trigger_require_approval_vote1(self):
        "Test trigger requirement: approval vote with one value"
        return self._test_require_approval_vote1('org/project2',
                                                 'project2-job')

    def _test_require_approval_vote1(self, project, job):
        A = self.fake_gerrit.addFakeChange(project, 'master', 'A')
        # A comment event that we will keep submitting to trigger
        comment = A.addApproval('Code-Review', 2, username='nobody')
        self.fake_gerrit.addEvent(comment)
        self.waitUntilSettled()
        # No approval from Jenkins so should not be enqueued
        self.assertEqual(len(self.history), 0)

        # A -1 from jenkins should not cause it to be enqueued
        A.addApproval('Verified', -1, username='jenkins')
        self.fake_gerrit.addEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 0)

        # A +1 should allow it to be enqueued
        A.addApproval('Verified', 1, username='jenkins')
        self.fake_gerrit.addEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)
        self.assertEqual(self.history[0].name, job)

        tenant = self.scheds.first.sched.abide.tenants.get('tenant-one')
        self.assertEqual(len(tenant.layout.loading_errors), 1)
        self.assertEqual(tenant.layout.loading_errors[0].name,
                         'Gerrit require-approval Deprecation')
        self.assertEqual(tenant.layout.loading_errors[0].severity,
                         'warning')
        self.assertIn('require-approval',
                      tenant.layout.loading_errors[0].short_error)
        self.assertIn('require-approval',
                      tenant.layout.loading_errors[0].error)
        self.assertIsNotNone(tenant.layout.loading_errors[0].key.context)
        self.assertIsNotNone(tenant.layout.loading_errors[0].key.mark)
        self.assertIsNotNone(tenant.layout.loading_errors[0].key.error_text)


class TestRequirementsVote2(ZuulTestCase):
    """Requirements with a voting requirement"""

    tenant_config_file = 'config/requirements/vote2/main.yaml'

    def test_pipeline_require_approval_vote2(self):
        "Test pipeline requirement: approval vote with two values"
        return self._test_require_approval_vote2('org/project1',
                                                 'project1-job')

    def test_trigger_require_approval_vote2(self):
        "Test trigger requirement: approval vote with two values"
        return self._test_require_approval_vote2('org/project2',
                                                 'project2-job')

    def _test_require_approval_vote2(self, project, job):
        A = self.fake_gerrit.addFakeChange(project, 'master', 'A')
        # A comment event that we will keep submitting to trigger
        comment = A.addApproval('Code-Review', 2, username='nobody')
        self.fake_gerrit.addEvent(comment)
        self.waitUntilSettled()
        # No approval from Jenkins so should not be enqueued
        self.assertEqual(len(self.history), 0)

        # A -1 from jenkins should not cause it to be enqueued
        A.addApproval('Verified', -1, username='jenkins')
        self.fake_gerrit.addEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 0)

        # A -2 from jenkins should not cause it to be enqueued
        A.addApproval('Verified', -2, username='jenkins')
        self.fake_gerrit.addEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 0)

        # A +1 from jenkins should allow it to be enqueued
        A.addApproval('Verified', 1, username='jenkins')
        self.fake_gerrit.addEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)
        self.assertEqual(self.history[0].name, job)

        # A +2 from nobody should not cause it to be enqueued
        B = self.fake_gerrit.addFakeChange(project, 'master', 'B')
        # A comment event that we will keep submitting to trigger
        comment = B.addApproval('Code-Review', 2, username='nobody')
        self.fake_gerrit.addEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)

        # A +2 from jenkins should allow it to be enqueued
        B.addApproval('Verified', 2, username='jenkins')
        self.fake_gerrit.addEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 2)
        self.assertEqual(self.history[1].name, job)

        tenant = self.scheds.first.sched.abide.tenants.get('tenant-one')
        self.assertEqual(len(tenant.layout.loading_errors), 1)
        self.assertEqual(tenant.layout.loading_errors[0].name,
                         'Gerrit require-approval Deprecation')
        self.assertEqual(tenant.layout.loading_errors[0].severity,
                         'warning')
        self.assertIn('require-approval',
                      tenant.layout.loading_errors[0].short_error)
        self.assertIn('require-approval',
                      tenant.layout.loading_errors[0].error)
        self.assertIsNotNone(tenant.layout.loading_errors[0].key.context)
        self.assertIsNotNone(tenant.layout.loading_errors[0].key.mark)
        self.assertIsNotNone(tenant.layout.loading_errors[0].key.error_text)


class TestRequirementsState(ZuulTestCase):
    """Requirements with simple state requirement"""

    tenant_config_file = 'config/requirements/state/main.yaml'

    def test_pipeline_require_current_patchset(self):
        # Create two patchsets and let their tests settle out. Then
        # comment on first patchset and check that no additional
        # jobs are run.
        A = self.fake_gerrit.addFakeChange('current-project', 'master', 'A')
        self.fake_gerrit.addEvent(A.addApproval('Code-Review', 1))
        self.waitUntilSettled()
        A.addPatchset()
        self.fake_gerrit.addEvent(A.addApproval('Code-Review', 1))
        self.waitUntilSettled()

        self.assertEqual(len(self.history), 2)  # one job for each ps
        self.fake_gerrit.addEvent(A.getChangeCommentEvent(1))
        self.waitUntilSettled()

        # Assert no new jobs ran after event for old patchset.
        self.assertEqual(len(self.history), 2)

        # Make sure the same event on a new PS will trigger
        self.fake_gerrit.addEvent(A.getChangeCommentEvent(2))
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 3)

    def test_pipeline_require_open(self):
        A = self.fake_gerrit.addFakeChange('open-project', 'master', 'A',
                                           status='MERGED')
        self.fake_gerrit.addEvent(A.addApproval('Code-Review', 2))
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 0)

        B = self.fake_gerrit.addFakeChange('open-project', 'master', 'B')
        self.fake_gerrit.addEvent(B.addApproval('Code-Review', 2))
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)

    def test_pipeline_require_status(self):
        A = self.fake_gerrit.addFakeChange('status-project', 'master', 'A',
                                           status='MERGED')
        self.fake_gerrit.addEvent(A.addApproval('Code-Review', 2))
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 0)

        B = self.fake_gerrit.addFakeChange('status-project', 'master', 'B')
        self.fake_gerrit.addEvent(B.addApproval('Code-Review', 2))
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)

    def test_pipeline_require_wip(self):
        A = self.fake_gerrit.addFakeChange('wip-project', 'master', 'A')
        A.setWorkInProgress(True)
        self.fake_gerrit.addEvent(A.addApproval('Code-Review', 2))
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 0)

        B = self.fake_gerrit.addFakeChange('wip-project', 'master', 'B')
        self.fake_gerrit.addEvent(B.addApproval('Code-Review', 2))
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)


class TestRequirementsRejectUsername(ZuulTestCase):
    """Requirements with reject username requirement"""

    tenant_config_file = 'config/requirements/reject-username/main.yaml'

    def _test_require_reject_username(self, project, job):
        "Test negative username's match"
        # Should only trigger if Jenkins hasn't voted.
        # add in a change with no comments
        A = self.fake_gerrit.addFakeChange(project, 'master', 'A')
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 0)

        # add in a comment that will trigger
        self.fake_gerrit.addEvent(A.addApproval('Code-Review', 1,
                                                username='reviewer'))
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)
        self.assertEqual(self.history[0].name, job)

        # add in a comment from jenkins user which shouldn't trigger
        self.fake_gerrit.addEvent(A.addApproval('Verified', 1,
                                                username='jenkins'))
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)

        # Check future reviews also won't trigger as a 'jenkins' user has
        # commented previously
        self.fake_gerrit.addEvent(A.addApproval('Code-Review', 1,
                                                username='reviewer'))
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)

    def test_pipeline_reject_username(self):
        "Test negative pipeline requirement: no comment from jenkins"
        return self._test_require_reject_username('org/project1',
                                                  'project1-job')

    def test_trigger_reject_username(self):
        "Test negative trigger requirement: no comment from jenkins"
        return self._test_require_reject_username('org/project2',
                                                  'project2-job')


class TestRequirementsReject(ZuulTestCase):
    """Requirements with reject requirement"""

    tenant_config_file = 'config/requirements/reject/main.yaml'

    def _test_require_reject(self, project, job):
        "Test no approval matches a reject param"
        A = self.fake_gerrit.addFakeChange(project, 'master', 'A')
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 0)

        # First positive vote should not queue until jenkins has +1'd
        comment = A.addApproval('Verified', 1, username='reviewer_a')
        self.fake_gerrit.addEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 0)

        # Jenkins should put in a +1 which will also queue
        comment = A.addApproval('Verified', 1, username='jenkins')
        self.fake_gerrit.addEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)
        self.assertEqual(self.history[0].name, job)

        # Negative vote should not queue
        comment = A.addApproval('Verified', -1, username='reviewer_b')
        self.fake_gerrit.addEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)

        # Future approvals should do nothing
        comment = A.addApproval('Verified', 1, username='reviewer_c')
        self.fake_gerrit.addEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)

        # Change/update negative vote should queue
        comment = A.addApproval('Verified', 1, username='reviewer_b')
        self.fake_gerrit.addEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 2)
        self.assertEqual(self.history[1].name, job)

        # Future approvals should also queue
        comment = A.addApproval('Verified', 1, username='reviewer_d')
        self.fake_gerrit.addEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 3)
        self.assertEqual(self.history[2].name, job)

    def test_pipeline_require_reject(self):
        "Test pipeline requirement: rejections absent"
        return self._test_require_reject('org/project1', 'project1-job')

    def test_trigger_require_reject(self):
        "Test trigger requirement: rejections absent"
        return self._test_require_reject('org/project2', 'project2-job')

    def test_pipeline_requirement_reject_unrelated(self):
        "Test if reject is obeyed if another unrelated approval is present"

        # Having no approvals whatsoever shall not reject the change
        A = self.fake_gerrit.addFakeChange('org/project3', 'master', 'A')
        self.fake_gerrit.addEvent(A.getChangeCommentEvent(1))
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)

        # Setting another unrelated approval shall not change the behavior of
        # the configured reject.
        comment = A.addApproval('Approved', 1, username='reviewer_e')
        self.fake_gerrit.addEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 2)

        # Setting the approval 'Verified' to a rejected value shall not lead to
        # a build.
        comment = A.addApproval('Verified', -1, username='jenkins')
        self.fake_gerrit.addEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 2)

        # Setting the approval 'Verified' to an accepted value shall lead to
        # a build.
        comment = A.addApproval('Verified', 1, username='jenkins')
        self.fake_gerrit.addEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 3)


class TestRequirementsTrustedCheck(ZuulTestCase):
    config_file = "zuul-gerrit-github.conf"
    tenant_config_file = "config/requirements/trusted-check/main.yaml"

    def test_non_live_requirements(self):
        # Test that pipeline requirements are applied to non-live
        # changes.
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        B = self.fake_gerrit.addFakeChange('org/project', 'master', 'B')
        B.setDependsOn(A, 1)
        B.addApproval('Code-Review', 2)

        self.fake_gerrit.addEvent(B.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()
        self.assertHistory([])

        self.fake_gerrit.addEvent(A.addApproval('Code-Review', 2))
        self.fake_gerrit.addEvent(B.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()
        self.assertHistory([
            dict(name='check-job', result='SUCCESS', changes='1,1 2,1')],
            ordered=False)

    def test_other_connections(self):
        # Test allow-other-connections: False
        A = self.fake_github.openFakePullRequest("gh/project", "master", "A")
        B = self.fake_gerrit.addFakeChange('org/project', 'master', 'B')
        B.data["commitMessage"] = "{}\n\nDepends-On: {}\n".format(
            B.subject, A.url,
        )
        B.addApproval('Code-Review', 2)

        self.fake_gerrit.addEvent(B.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()
        self.assertHistory([])


class TestGerritTriggerRequirements(ZuulTestCase):
    scheduler_count = 1

    @simple_layout('layouts/gerrit-trigger-requirements.yaml')
    def test_require_open(self):
        # Test trigger require-open
        jobname = 'require-open'
        project = 'org/project'
        A = self.fake_gerrit.addFakeChange(project, 'master', 'A')
        # A comment event that we will keep submitting to trigger
        comment = A.getChangeCommentEvent(1, f'test {jobname}')

        # It's open, so it should be enqueued
        self.fake_gerrit.addEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)
        self.assertEqual(self.history[0].name, jobname)

        # Not open, so should be ignored
        A.setMerged()
        self.fake_gerrit.addEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)
        self.assertEqual(self.history[0].name, jobname)

    @simple_layout('layouts/gerrit-trigger-requirements.yaml')
    def test_reject_open(self):
        # Test trigger reject-open
        jobname = 'reject-open'
        project = 'org/project'
        A = self.fake_gerrit.addFakeChange(project, 'master', 'A')
        # A comment event that we will keep submitting to trigger
        comment = A.getChangeCommentEvent(1, f'test {jobname}')

        # It's open, so it should not be enqueued
        self.fake_gerrit.addEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 0)

        # Not open, so should be enqueued
        A.setMerged()
        self.fake_gerrit.addEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)
        self.assertEqual(self.history[0].name, jobname)

    @simple_layout('layouts/gerrit-trigger-requirements.yaml')
    def test_require_wip(self):
        # Test trigger require-wip
        jobname = 'require-wip'
        project = 'org/project'
        A = self.fake_gerrit.addFakeChange(project, 'master', 'A')
        # A comment event that we will keep submitting to trigger
        comment = A.getChangeCommentEvent(1, f'test {jobname}')

        # It's not WIP, so it should be ignored
        self.fake_gerrit.addEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 0)

        # WIP, so should be enqueued
        A.setWorkInProgress(True)
        self.fake_gerrit.addEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)
        self.assertEqual(self.history[0].name, jobname)

    @simple_layout('layouts/gerrit-trigger-requirements.yaml')
    def test_reject_wip(self):
        # Test trigger reject-wip
        jobname = 'reject-wip'
        project = 'org/project'
        A = self.fake_gerrit.addFakeChange(project, 'master', 'A')
        # A comment event that we will keep submitting to trigger
        comment = A.getChangeCommentEvent(1, f'test {jobname}')

        # It's not WIP, so it should be enqueued
        self.fake_gerrit.addEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)
        self.assertEqual(self.history[0].name, jobname)

        # WIP, so should be ignored
        A.setWorkInProgress(True)
        self.fake_gerrit.addEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)
        self.assertEqual(self.history[0].name, jobname)

    @simple_layout('layouts/gerrit-trigger-requirements.yaml')
    def test_require_current_patchset(self):
        # Test trigger require-current_patchset
        jobname = 'require-current-patchset'
        project = 'org/project'
        A = self.fake_gerrit.addFakeChange(project, 'master', 'A')
        # A comment event that we will keep submitting to trigger
        comment = A.getChangeCommentEvent(1, f'test {jobname}')

        # It's current, so it should be enqueued
        self.fake_gerrit.addEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)
        self.assertEqual(self.history[0].name, jobname)

        # Not current, so should be ignored
        A.addPatchset()
        self.fake_gerrit.addEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)
        self.assertEqual(self.history[0].name, jobname)

    @simple_layout('layouts/gerrit-trigger-requirements.yaml')
    def test_reject_current_patchset(self):
        # Test trigger reject-current_patchset
        jobname = 'reject-current-patchset'
        project = 'org/project'
        A = self.fake_gerrit.addFakeChange(project, 'master', 'A')
        # A comment event that we will keep submitting to trigger
        comment = A.getChangeCommentEvent(1, f'test {jobname}')

        # It's current, so it should be ignored
        self.fake_gerrit.addEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 0)

        # Not current, so should be enqueued
        A.addPatchset()
        self.fake_gerrit.addEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)
        self.assertEqual(self.history[0].name, jobname)

    @simple_layout('layouts/gerrit-trigger-requirements.yaml')
    def test_require_status(self):
        # Test trigger require-status
        jobname = 'require-status'
        project = 'org/project'
        A = self.fake_gerrit.addFakeChange(project, 'master', 'A')
        # A comment event that we will keep submitting to trigger
        comment = A.getChangeCommentEvent(1, f'test {jobname}')

        # It's not merged, so it should be ignored
        self.fake_gerrit.addEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 0)

        # Merged, so should be enqueued
        A.setMerged()
        self.fake_gerrit.addEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)
        self.assertEqual(self.history[0].name, jobname)

    @simple_layout('layouts/gerrit-trigger-requirements.yaml')
    def test_reject_status(self):
        # Test trigger reject-status
        jobname = 'reject-status'
        project = 'org/project'
        A = self.fake_gerrit.addFakeChange(project, 'master', 'A')
        # A comment event that we will keep submitting to trigger
        comment = A.getChangeCommentEvent(1, f'test {jobname}')

        # It's not merged, so it should be enqueued
        self.fake_gerrit.addEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)
        self.assertEqual(self.history[0].name, jobname)

        # Merged, so should be ignored
        A.setMerged()
        self.fake_gerrit.addEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)
        self.assertEqual(self.history[0].name, jobname)

    @simple_layout('layouts/gerrit-trigger-requirements.yaml')
    def test_require_approval(self):
        # Test trigger require-approval
        jobname = 'require-approval'
        project = 'org/project'
        A = self.fake_gerrit.addFakeChange(project, 'master', 'A')
        # A comment event that we will keep submitting to trigger
        comment = A.getChangeCommentEvent(1, f'test {jobname}')

        # Missing approval, so it should be ignored
        self.fake_gerrit.addEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 0)

        # Has approval, so it should be enqueued
        A.addApproval('Verified', 1, username='zuul')
        self.fake_gerrit.addEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)
        self.assertEqual(self.history[0].name, jobname)

    @simple_layout('layouts/gerrit-trigger-requirements.yaml')
    def test_reject_approval(self):
        # Test trigger reject-approval
        jobname = 'reject-approval'
        project = 'org/project'
        A = self.fake_gerrit.addFakeChange(project, 'master', 'A')
        # A comment event that we will keep submitting to trigger
        comment = A.getChangeCommentEvent(1, f'test {jobname}')

        # Missing approval, so it should be enqueued
        self.fake_gerrit.addEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)
        self.assertEqual(self.history[0].name, jobname)

        # Has approval, so it should be ignored
        A.addApproval('Verified', 1, username='zuul')
        self.fake_gerrit.addEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)
        self.assertEqual(self.history[0].name, jobname)


class TestPromoteWithRequirements(ZuulTestCase):
    @simple_layout('layouts/gerrit-gate-requirements.yaml')
    def test_promote_all_matching_requirements(self):
        self.executor_server.hold_jobs_in_build = True
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        B = self.fake_gerrit.addFakeChange('org/project', 'master', 'B')
        C = self.fake_gerrit.addFakeChange('org/project', 'master', 'C')
        C.setDependsOn(B, 1)
        A.addApproval('Code-Review', 2)
        B.addApproval('Code-Review', 2)
        C.addApproval('Code-Review', 2)
        A.data['hashtags'] = ['oktomerge']
        B.data['hashtags'] = ['oktomerge']
        C.data['hashtags'] = ['oktomerge']
        self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
        # Do not emit an event for this to force processing of B to discover
        # the approved state
        C.addApproval('Approved', 1)
        self.fake_gerrit.addEvent(B.addApproval('Approved', 1))
        self.waitUntilSettled()

        # Before the promote we have A <- B <- C
        self.assertEqual(len(self.builds), 3)
        self.assertTrue(self.builds[0].hasChanges(A))
        self.assertTrue(self.builds[1].hasChanges(A, B))
        self.assertTrue(self.builds[2].hasChanges(A, B, C))

        # Promote change B to the head of the gate queue
        event = PromoteEvent(
            'tenant-one', 'gate', [f'{B.number},{B.latest_patchset}'])
        self.scheds.first.sched.pipeline_management_events['tenant-one'][
            'gate'].put(event)
        self.waitUntilSettled()

        # After the promote we have B <- C <- A
        self.assertEqual(len(self.builds), 3)
        self.assertTrue(self.builds[0].hasChanges(B))
        self.assertTrue(self.builds[1].hasChanges(B, C))
        self.assertTrue(self.builds[2].hasChanges(B, C, A))

        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()

        self.assertEqual(B.data['status'], 'MERGED')
        self.assertEqual(C.data['status'], 'MERGED')
        self.assertEqual(A.data['status'], 'MERGED')

    @simple_layout('layouts/gerrit-gate-requirements.yaml')
    def test_promote_child_not_matching_requirements(self):
        self.executor_server.hold_jobs_in_build = True
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        B = self.fake_gerrit.addFakeChange('org/project', 'master', 'B')
        C = self.fake_gerrit.addFakeChange('org/project', 'master', 'C')
        C.setDependsOn(B, 1)
        A.addApproval('Code-Review', 2)
        C.addApproval('Code-Review', 2)
        B.addApproval('Code-Review', 2)
        A.data['hashtags'] = ['oktomerge']
        B.data['hashtags'] = ['oktomerge']
        # The gate pipeline requires the "oktomerge" hashtag; we omit
        # it from change C to test that the pipeline requirement still
        # applies to child changes not already in the pipeline even
        # when promote is used.
        self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
        self.fake_gerrit.addEvent(B.addApproval('Approved', 1))
        self.fake_gerrit.addEvent(C.addApproval('Approved', 1))
        self.waitUntilSettled()

        # Before the promote we have A <- B
        self.assertEqual(len(self.builds), 2)
        self.assertTrue(self.builds[0].hasChanges(A))
        self.assertTrue(self.builds[1].hasChanges(A, B))

        # Promote change B to the head of the gate queue
        event = PromoteEvent(
            'tenant-one', 'gate', [f'{B.number},{B.latest_patchset}'])
        self.scheds.first.sched.pipeline_management_events['tenant-one'][
            'gate'].put(event)
        self.waitUntilSettled()

        # After the promote we have B <- A
        # Promoting a change that meets requirements should succeed
        # but we shouldn't enqueue its children that do not meet
        # requirements.
        self.assertEqual(len(self.builds), 2)
        self.assertTrue(self.builds[0].hasChanges(B))
        self.assertTrue(self.builds[1].hasChanges(B, A))

        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()

        self.assertEqual(B.data['status'], 'MERGED')
        self.assertEqual(A.data['status'], 'MERGED')
        self.assertEqual(C.data['status'], 'NEW')

    @simple_layout('layouts/gerrit-gate-requirements.yaml')
    @okay_tracebacks('Unable to find 3,1')
    def test_promote_parent_not_matching_requirements(self):
        self.executor_server.hold_jobs_in_build = True
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        B = self.fake_gerrit.addFakeChange('org/project', 'master', 'B')
        C = self.fake_gerrit.addFakeChange('org/project', 'master', 'C')
        C.setDependsOn(B, 1)
        A.addApproval('Code-Review', 2)
        B.addApproval('Code-Review', 2)
        C.addApproval('Code-Review', 2)
        A.data['hashtags'] = ['oktomerge']
        # The gate pipeline requires the "oktomerge" hashtag; we omit
        # it from change B to test that the pipeline requirement still
        # applies to parent changes not already in the pipeline even
        # when promote is used.
        C.data['hashtags'] = ['oktomerge']
        self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
        self.fake_gerrit.addEvent(B.addApproval('Approved', 1))
        self.fake_gerrit.addEvent(C.addApproval('Approved', 1))
        self.waitUntilSettled()

        # Before the promote we have A
        self.assertEqual(len(self.builds), 1)
        self.assertTrue(self.builds[0].hasChanges(A))
        build = self.builds[0]

        # Promote change C to the head of the gate queue
        event = PromoteEvent(
            'tenant-one', 'gate', [f'{C.number},{C.latest_patchset}'])
        self.scheds.first.sched.pipeline_management_events['tenant-one'][
            'gate'].put(event)
        self.waitUntilSettled()

        # After the promote we still have just A because C cannot enqueue
        # due to its parent missing requirements.
        self.assertEqual(len(self.builds), 1)
        self.assertTrue(self.builds[0].hasChanges(A))
        self.assertEqual(build, self.builds[0])

        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()

        self.assertEqual(A.data['status'], 'MERGED')
        self.assertEqual(B.data['status'], 'NEW')
        self.assertEqual(C.data['status'], 'NEW')

    @simple_layout('layouts/gerrit-gate-requirements.yaml')
    @okay_tracebacks('Unable to find 3,1')
    def test_promote_not_matching_requirements(self):
        self.executor_server.hold_jobs_in_build = True
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        B = self.fake_gerrit.addFakeChange('org/project', 'master', 'B')
        C = self.fake_gerrit.addFakeChange('org/project', 'master', 'C')
        C.setDependsOn(B, 1)
        A.addApproval('Code-Review', 2)
        B.addApproval('Code-Review', 2)
        C.addApproval('Code-Review', 2)
        A.data['hashtags'] = ['oktomerge']
        B.data['hashtags'] = ['oktomerge']
        # The gate pipeline requires the "oktomerge" hashtag; we omit
        # it from change C to test that the pipeline requirement still
        # applies to changes that are directly promoted.
        self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
        self.fake_gerrit.addEvent(B.addApproval('Approved', 1))
        self.fake_gerrit.addEvent(C.addApproval('Approved', 1))
        self.waitUntilSettled()

        # Before the promote we have A <- B
        self.assertEqual(len(self.builds), 2)
        self.assertTrue(self.builds[0].hasChanges(A))
        self.assertTrue(self.builds[1].hasChanges(A, B))
        builds = (self.builds[0], self.builds[1])

        # Attempt to promote change C to the head of the gate queue.
        # C isn't in the pipeline and doesn't meet requirements. It will
        # fail to promote and so will its parent. This is a noop.
        event = PromoteEvent(
            'tenant-one', 'gate', [f'{C.number},{C.latest_patchset}'])
        self.scheds.first.sched.pipeline_management_events['tenant-one'][
            'gate'].put(event)
        self.waitUntilSettled()

        # After promotion we still have A <- B because promoting C
        # will fail as it was never in the pipeline in the first place.
        # Attempting to promote C produces the error we accept as ok
        # via the test case decorator.
        self.assertEqual(len(self.builds), 2)
        self.assertTrue(self.builds[0].hasChanges(A))
        self.assertTrue(self.builds[1].hasChanges(A, B))
        self.assertEqual(builds, (self.builds[0], self.builds[1]))

        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()

        self.assertEqual(A.data['status'], 'MERGED')
        self.assertEqual(B.data['status'], 'MERGED')
        self.assertEqual(C.data['status'], 'NEW')

    @simple_layout('layouts/gerrit-gate-requirements.yaml')
    def test_promote_unrelated_not_matching_requirements(self):
        self.executor_server.hold_jobs_in_build = True
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        B = self.fake_gerrit.addFakeChange('org/project', 'master', 'B')
        C = self.fake_gerrit.addFakeChange('org/project', 'master', 'C')
        C.setDependsOn(B, 1)
        A.addApproval('Code-Review', 2)
        B.addApproval('Code-Review', 2)
        C.addApproval('Code-Review', 2)
        # The gate pipeline requires the "oktomerge" hashtag; we omit
        # it from change A to test that trivial case that changes not meeting
        # pipeline requirements are not added to the queue if other unrelated
        # changes are directly promoted.
        B.data['hashtags'] = ['oktomerge']
        C.data['hashtags'] = ['oktomerge']
        self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
        self.fake_gerrit.addEvent(B.addApproval('Approved', 1))
        self.fake_gerrit.addEvent(C.addApproval('Approved', 1))
        self.waitUntilSettled()

        # Before the promote we have B <- C
        self.assertEqual(len(self.builds), 2)
        self.assertTrue(self.builds[0].hasChanges(B))
        self.assertTrue(self.builds[1].hasChanges(B, C))

        # Attempt to promote change C to the head of the gate queue.
        # A has no relationship to B or C and should be ignored.
        # C depends on B so will not jump ahead of B in the queue.
        event = PromoteEvent(
            'tenant-one', 'gate', [f'{C.number},{C.latest_patchset}'])
        self.scheds.first.sched.pipeline_management_events['tenant-one'][
            'gate'].put(event)
        self.waitUntilSettled()

        # After promotion we have B <- C
        self.assertEqual(len(self.builds), 2)
        self.assertTrue(self.builds[0].hasChanges(B))
        self.assertTrue(self.builds[1].hasChanges(B, C))

        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()

        self.assertEqual(B.data['status'], 'MERGED')
        self.assertEqual(C.data['status'], 'MERGED')
        self.assertEqual(A.data['status'], 'NEW')



# Copyright 2020 OpenStack Foundation
# Copyright 2020 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
from unittest import mock
import os.path
import jwt
from io import StringIO
import time

from zuul.driver import auth

from tests.base import BaseTestCase, FIXTURE_DIR

with open(os.path.join(FIXTURE_DIR,
                       'auth/openid-configuration.json'), 'r') as well_known:
    FAKE_WELL_KNOWN_CONFIG = json.loads(well_known.read())


algo = jwt.algorithms.RSAAlgorithm(jwt.algorithms.RSAAlgorithm.SHA256)
with open(os.path.join(FIXTURE_DIR,
                       'auth/oidc-key'), 'r') as k:
    OIDC_PRIVATE_KEY = algo.prepare_key(k.read())
with open(os.path.join(FIXTURE_DIR,
                       'auth/oidc-key.pub'), 'r') as k:
    pub_key = algo.prepare_key(k.read())
    pub_jwk = algo.to_jwk(pub_key)
    key = {
        "kid": "OwO",
        "use": "sig",
        "alg": "RS256"
    }
    key.update(json.loads(pub_jwk))
    # not present in keycloak jwks
    if "key_ops" in key:
        del key["key_ops"]
    FAKE_CERTS = {
        "keys": [
            key
        ]
    }


class FakeResponse:
    def __init__(self, json_dict):
        self._json = json_dict

    def json(self):
        return self._json


def mock_get(url, params=None, **kwargs):
    if url == ("https://my.oidc.provider/auth/realms/realm-one/"
               ".well-known/openid-configuration"):
        return FakeResponse(FAKE_WELL_KNOWN_CONFIG)
    else:
        raise Exception("Unknown URL %s" % url)


def mock_urlopen(url, *args, **kwargs):
    if hasattr(url, 'full_url'):
        # Like a urllib.Request object
        url = url.full_url
    if url == ("https://my.oidc.provider/auth/realms/realm-one/"
               "protocol/openid-connect/certs"):
        io = StringIO()
        json.dump(FAKE_CERTS, io)
        io.seek(0)
        return io
    else:
        raise Exception("Unknown URL %s" % url)


class TestOpenIDConnectAuthenticator(BaseTestCase):
    def test_decodeToken(self):
        """Test the decoding workflow"""
        config = {
            'issuer_id': FAKE_WELL_KNOWN_CONFIG['issuer'],
            'client_id': 'zuul-app',
            'realm': 'realm-one',
        }
        OIDCAuth = auth.jwt.OpenIDConnectAuthenticator(**config)
        payload = {
            'iss': FAKE_WELL_KNOWN_CONFIG['issuer'],
            'aud': config['client_id'],
            'exp': int(time.time()) + 3600,
            'sub': 'someone'
        }
        token = jwt.encode(
            payload,
            OIDC_PRIVATE_KEY,
            algorithm='RS256',
            headers={'kid': 'OwO'})
        with mock.patch('requests.get', side_effect=mock_get):
            # patching call in PyJWKClient's fetch_data
            with mock.patch('urllib.request.urlopen',
                            side_effect=mock_urlopen):
                decoded = OIDCAuth.decodeToken(token)
                for claim in payload.keys():
                    self.assertEqual(payload[claim], decoded[claim])



# Copyright 2024-2025 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import os

from zuul.connection import BaseConnection
from zuul.lib.config import any_to_bool


class KubernetesConnection(BaseConnection):
    driver_name = 'kubernetes'
    log = logging.getLogger("zuul.KubernetesConnection")

    def __init__(self, driver, connection_name, connection_config):
        super().__init__(driver, connection_name, connection_config)

        # Allow the user to specify the kubeconfig via the config
        # setting or the environment variable.  Otherwise we will use
        # the client lib default paths.
        self.config_file = self.connection_config.get(
            'kubeconfig_file',
            os.getenv('KUBECONFIG', None))

        self.context = self.connection_config.get('context')

        # Rate limit: requests/second
        self.rate = self.connection_config.get('rate', 2)
        # Use openshift projects instead of namespaces
        self.use_openshift_projects = any_to_bool(self.connection_config.get(
            'use_openshift_projects', False))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""fix_branchless_refs

Revision ID: d0a269345a22
Revises: f7843ddf1552
Create Date: 2023-11-21 12:04:22.123186

"""

# revision identifiers, used by Alembic.
revision = 'd0a269345a22'
down_revision = 'f7843ddf1552'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


REF_TABLE = 'zuul_ref'


def upgrade(table_prefix=''):
    prefixed_ref = table_prefix + REF_TABLE
    statement = f"""
        update {prefixed_ref}
        set branch=substring(ref from 12)
        where branch='' and ref like 'refs/heads/%'
    """
    connection = op.get_bind()
    connection.execute(sa.text(statement))


def downgrade():
    raise Exception("Downgrades not supported")



# Copyright 2018 Red Hat, Inc.
# Copyright 2024 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import time
import logging

import voluptuous as v

from zuul.reporter import BaseReporter
from zuul.exceptions import MergeFailure
from zuul.driver.pagure.paguresource import PagureSource


class PagureReporter(BaseReporter):
    """Sends off reports to Pagure."""

    name = 'pagure'
    log = logging.getLogger("zuul.PagureReporter")

    def __init__(self, driver, connection, pipeline, config=None,
                 parse_context=None):
        super(PagureReporter, self).__init__(driver, connection, config,
                                             parse_context)
        self._commit_status = self.config.get('status', None)
        self._create_comment = self.config.get('comment', True)
        self._merge = self.config.get('merge', False)

    def getContext(self, manager):
        return "{}/{}".format(manager.tenant.name,
                              manager.pipeline.name)

    def report(self, item, phase1=True, phase2=True):
        """Report on an event."""
        for change in item.changes:
            self._reportChange(item, change, phase1, phase2)
        return []

    def _reportChange(self, item, change, phase1=True, phase2=True):
        """Report on an event."""

        # If the source is not PagureSource we cannot report anything here.
        if not isinstance(change.project.source, PagureSource):
            return

        # For supporting several Pagure connections we also must filter by
        # the canonical hostname.
        if change.project.source.connection.canonical_hostname != \
                self.connection.canonical_hostname:
            return

        if phase1:
            if self._commit_status is not None:
                if (hasattr(change, 'patchset') and
                        change.patchset is not None):
                    self.setCommitStatus(item, change)
                elif (hasattr(change, 'newrev') and
                        change.newrev is not None):
                    self.setCommitStatus(item, change)
            if hasattr(change, 'number'):
                if self._create_comment:
                    self.addPullComment(item, change)
        if phase2 and self._merge:
            self.mergePull(item, change)
            if not change.is_merged:
                msg = self._formatItemReportMergeConflict(item, change)
                self.addPullComment(item, change, msg)

    def _formatItemReportJobs(self, item, short):
        # Return the list of jobs portion of the report
        ret = ''
        jobs_fields, skipped = self._getItemReportJobsFields(item)
        for job_fields in jobs_fields:
            ret += '- [%s](%s) : %s%s%s%s\n' % job_fields[:6]
        if skipped:
            jobtext = 'job' if skipped == 1 else 'jobs'
            ret += 'Skipped %i %s\n' % (skipped, jobtext)
        return ret

    def addPullComment(self, item, change, comment=None):
        message = comment or self._formatItemReport(item)
        project = change.project.name
        pr_number = change.number
        self.log.debug(
            'Reporting change %s, params %s, message: %s' %
            (change, self.config, message))
        self.connection.commentPull(project, pr_number, message)

    def setCommitStatus(self, item, change):
        project = change.project.name
        if hasattr(change, 'patchset'):
            sha = change.patchset
        elif hasattr(change, 'newrev'):
            sha = change.newrev
        state = self._commit_status
        change_number = change.number

        url = item.formatItemUrl()
        description = '%s status: %s (%s)' % (
            item.manager.pipeline.name, self._commit_status, sha)

        self.log.debug(
            'Reporting change %s, params %s, '
            'context: %s, state: %s, description: %s, url: %s' %
            (change, self.config,
             self.getContext(item.manager), state, description, url))

        self.connection.setCommitStatus(
            project, change_number, state, url, description,
            self.getContext(item.manager))

    def mergePull(self, item, change):
        project = change.project.name
        pr_number = change.number

        for i in [1, 2]:
            try:
                self.connection.mergePull(project, pr_number)
                change.is_merged = True
                return
            except MergeFailure:
                self.log.exception(
                    'Merge attempt of change %s  %s/2 failed.' %
                    (change, i), exc_info=True)
                if i == 1:
                    time.sleep(2)
        self.log.warning(
            'Merge of change %s failed after 2 attempts, giving up' %
            change)

    def getSubmitAllowNeeds(self, manager):
        return []


def getSchema():
    pagure_reporter = v.Schema({
        'status': v.Any('pending', 'success', 'failure'),
        'comment': bool,
        'merge': bool,
    })
    return pagure_reporter



#!/usr/bin/env python

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os

from zuul.lib import yamlutil as yaml

FIXTURE_DIR = os.path.join(os.path.dirname(__file__),
                           'fixtures')
CONFIG_DIR = os.path.join(FIXTURE_DIR, 'config')


def make_playbook(path):
    d = os.path.dirname(path)
    try:
        os.makedirs(d)
    except OSError:
        pass
    with open(path, 'w') as f:
        f.write('- hosts: all\n')
        f.write('  tasks: []\n')


def handle_repo(path):
    print('Repo: %s' % path)
    config_path = None
    for fn in ['zuul.yaml', '.zuul.yaml']:
        if os.path.exists(os.path.join(path, fn)):
            config_path = os.path.join(path, fn)
            break
    try:
        with open(config_path) as f:
            config = yaml.safe_load(f)
    except Exception:
        print("  Has yaml errors")
        return
    for block in config:
        if 'job' not in block:
            continue
        job = block['job']['name']
        playbook = os.path.join(path, 'playbooks', job + '.yaml')
        if not os.path.exists(playbook):
            print('  Creating: %s' % job)
            make_playbook(playbook)


def main():
    repo_dirs = []

    for root, dirs, files in os.walk(CONFIG_DIR):
        if 'zuul.yaml' in files or '.zuul.yaml' in files:
            repo_dirs.append(root)

    for path in repo_dirs:
        handle_repo(path)


if __name__ == '__main__':
    main()



# Copyright 2025 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from enum import StrEnum
import json
import logging
import math
import os
import urllib.parse

from kazoo.exceptions import (
    BadVersionError,
    NodeExistsError,
)

from zuul import model
from zuul.zk.cache import ZuulTreeCache
from zuul.zk.zkobject import ZKObject, ZKContext


class ZKQuotaInfo(ZKObject):
    def __init__(self):
        super().__init__()

    def getPath(self):
        path = f'zuul/endpoint/{self.endpoint}/quota/{self.kind}'
        if self.resource is not None:
            path += f'/{self.resource}'
        return path

    def serialize(self, context):
        data = {
            'quota': self.quota,
        }
        return json.dumps(data, sort_keys=True).encode("utf8")


class QuotaCache(ZuulTreeCache):
    """Stores endpoint quota information in ZK

    This stores two types of information: the overall quota limits of
    the endpoint, and resource usage for instance types, etc.

    The overall quota usage is expected to change very infrequently,
    and the resource usage even less (never).  To that end, any writes
    that fail due to concurrent modifications are simply ignored under
    the assumption that another launcher is refreshing the data at the
    same time.
    """

    log = logging.getLogger('zuul.QuotaCache')

    class Kind(StrEnum):
        LIMITS = "limits"
        UNMANAGED_USAGE = "unmanaged-usage"
        RESOURCE = "resource"

    def __init__(self, client, endpoint_name):
        self.endpoint = urllib.parse.quote_plus(endpoint_name)
        root = f'/zuul/endpoint/{self.endpoint}/quota'
        resource_path = os.path.join(root, QuotaCache.Kind.RESOURCE)
        super().__init__(client, root, async_worker=False)
        self.zk_client.client.ensure_path(resource_path)

    def objectFromRaw(self, key, data, zstat):
        if len(key) == 2:
            resource = key[1]
        else:
            resource = None
        obj = ZKQuotaInfo._fromRaw(self._zk_context, data, zstat, None)
        obj._set(endpoint=self.endpoint,
                 kind=key[0],
                 resource=resource)
        return obj

    def updateFromRaw(self, obj, key, data, zstat):
        obj._updateFromRaw(self._zk_context, data, zstat, None)

    def _makeKey(self, kind, resource=None):
        if kind == QuotaCache.Kind.RESOURCE:
            return (str(kind), resource)
        return (str(kind),)

    def parsePath(self, path):
        parts = path.split('/')
        parts = parts[5:]
        key = None
        if len(parts) == 1 and parts[0] == QuotaCache.Kind.LIMITS:
            key = self._makeKey(QuotaCache.Kind.LIMITS)
        elif len(parts) == 1 and parts[0] == QuotaCache.Kind.UNMANAGED_USAGE:
            key = self._makeKey(QuotaCache.Kind.UNMANAGED_USAGE)
        elif len(parts) == 2 and parts[0] == QuotaCache.Kind.RESOURCE:
            key = self._makeKey(QuotaCache.Kind.RESOURCE, parts[1])
        # We should fetch if we have a matching key
        return (key, bool(key))

    def hasLimits(self):
        key = self._makeKey(QuotaCache.Kind.LIMITS)
        obj = self._cached_objects.get(key)
        return bool(obj is not None)

    def hasUnmanagedUsage(self):
        key = self._makeKey(QuotaCache.Kind.UNMANAGED_USAGE)
        obj = self._cached_objects.get(key)
        return bool(obj is not None)

    def hasResource(self, resource):
        key = self._makeKey(QuotaCache.Kind.RESOURCE, resource)
        obj = self._cached_objects.get(key)
        return bool(obj is not None)

    def getLimits(self):
        key = self._makeKey(QuotaCache.Kind.LIMITS)
        obj = self._cached_objects.get(key)
        quota = obj and obj.quota or {}
        return model.QuotaInformation(default=math.inf, **quota)

    def getUnmanagedUsage(self):
        key = self._makeKey(QuotaCache.Kind.UNMANAGED_USAGE)
        obj = self._cached_objects.get(key)
        quota = obj and obj.quota or {}
        return model.QuotaInformation(**quota)

    def getResource(self, resource):
        key = self._makeKey(QuotaCache.Kind.RESOURCE, resource)
        obj = self._cached_objects.get(key)
        if obj is None:
            raise KeyError(f"Resource {resource} not in quota cache")
        return model.QuotaInformation(**obj.quota)

    def createZKContext(self):
        return ZKContext(self.zk_client, None, None, self.log)

    def _setValue(self, kind, resource, quota_info):
        path = os.path.join(self.root, kind)
        if kind == QuotaCache.Kind.RESOURCE:
            key = (kind, resource)
        else:
            key = (kind,)
        obj = self._cached_objects.get(key)
        context = self.createZKContext()
        try:
            if obj:
                obj.updateAttributes(context, quota=quota_info.quota)
            else:
                obj = ZKQuotaInfo.new(
                    context,
                    quota=quota_info.quota,
                    endpoint=self.endpoint,
                    kind=kind,
                    resource=resource,
                )
                # Put the item in the cache so that we can use it
                # immediately without hitting a cache error.  This
                # will race the addition triggered by the watch, but
                # we'll end up with the same data, and no consumers of
                # the quota cache would be affected by having the
                # object replaced.
                self._cached_objects.setdefault(key, obj)
        except (BadVersionError, NodeExistsError):
            self.log.debug("Skipping update of %s", path)

    def setLimits(self, quota_info):
        self._setValue(QuotaCache.Kind.LIMITS, None, quota_info)

    def setUnmanagedUsage(self, quota_info):
        self._setValue(QuotaCache.Kind.UNMANAGED_USAGE, None, quota_info)

    def setResource(self, resource, quota_info):
        self._setValue(QuotaCache.Kind.RESOURCE, resource, quota_info)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import fixtures
import logging
import subprocess
import sys
import tempfile
import testtools
import time
import os

from zuul.driver import bubblewrap
from zuul.executor.server import SshAgent
from tests.base import iterate_timeout
from unittest import skipIf


class TestBubblewrap(testtools.TestCase):
    def setUp(self):
        super(TestBubblewrap, self).setUp()
        self.log_fixture = self.useFixture(
            fixtures.FakeLogger(level=logging.DEBUG))
        self.useFixture(fixtures.NestedTempfile())

    def test_bubblewrap_wraps(self):
        bwrap = bubblewrap.BubblewrapDriver(check_bwrap=True)
        context = bwrap.getExecutionContext()
        work_dir = tempfile.mkdtemp()
        ssh_agent = SshAgent()
        self.addCleanup(ssh_agent.stop)
        ssh_agent.start()
        po = context.getPopen(work_dir=work_dir,
                              ssh_auth_sock=ssh_agent.env['SSH_AUTH_SOCK'])
        self.assertTrue(po.fds[0] > 2)
        self.assertTrue(po.fds[1] > 2)
        self.assertTrue(work_dir in po.command)
        # Now run /usr/bin/id to verify passwd/group entries made it in
        true_proc = po(['/usr/bin/id'], stdout=subprocess.PIPE,
                       stderr=subprocess.PIPE)
        (output, errs) = true_proc.communicate()
        # Make sure it printed things on stdout
        self.assertTrue(len(output.strip()))
        # And that it did not print things on stderr
        self.assertEqual(0, len(errs.strip()))
        # Make sure the _r's are closed
        self.assertEqual([], po.fds)

    @skipIf(sys.platform == 'darwin', 'Not supported on MacOS')
    def test_bubblewrap_leak(self):
        bwrap = bubblewrap.BubblewrapDriver(check_bwrap=True)
        context = bwrap.getExecutionContext()
        work_dir = tempfile.mkdtemp()
        ansible_dir = tempfile.mkdtemp()
        ssh_agent = SshAgent()
        self.addCleanup(ssh_agent.stop)
        ssh_agent.start()
        po = context.getPopen(work_dir=work_dir,
                              ansible_dir=ansible_dir,
                              ssh_auth_sock=ssh_agent.env['SSH_AUTH_SOCK'])
        leak_time = 60
        # Use hexadecimal notation to avoid false-positive
        true_proc = po(['bash', '-c', 'sleep 0x%X & disown' % leak_time])
        self.assertEqual(0, true_proc.wait())
        cmdline = "sleep\x000x%X\x00" % leak_time
        for x in iterate_timeout(30, "process to exit"):
            try:
                sleep_proc = []
                for pid in os.listdir("/proc"):
                    if os.path.isfile("/proc/%s/cmdline" % pid):
                        with open("/proc/%s/cmdline" % pid) as f:
                            if f.read() == cmdline:
                                sleep_proc.append(pid)

                if not sleep_proc:
                    break
            except FileNotFoundError:
                pass
            except ProcessLookupError:
                pass
            time.sleep(1)



#!/usr/bin/env python3

# Copyright 2020 BMW Group
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import argparse
import os
import subprocess


def main():
    pos_args = {
        '--dir': 1,
        '--tmpfs': 1,
        '--ro-bind': 2,
        '--bind': 2,
        '--chdir': 1,
        '--uid': 1,
        '--gid': 1,
        '--file': 2,
        '--proc': 1,
        '--dev': 1,
    }
    bool_args = [
        '--unshare-all',
        '--unshare-user',
        '--unshare-user-try',
        '--unshare-ipc',
        '--unshare-pid',
        '--unshare-net',
        '--unshare-uts',
        '--unshare-cgroup',
        '--unshare-cgroup-try',
        '--share-net',
        '--die-with-parent',
    ]
    parser = argparse.ArgumentParser()
    for arg, nargs in pos_args.items():
        parser.add_argument(arg, nargs=nargs, action='append')
    for arg in bool_args:
        parser.add_argument(arg, action='store_true')
    parser.add_argument('args', metavar='args', nargs=argparse.REMAINDER,
                        help='Command')

    args = parser.parse_args()

    for fd, path in args.file:
        fd = int(fd)
        if path.startswith('/etc'):
            # Ignore write requests to /etc
            continue
        print('Writing file from %s to %s' % (fd, path))
        count = 0
        with open(path, 'wb') as output:
            data = os.read(fd, 32000)
            while data:
                count += len(data)
                output.write(data)
                data = os.read(fd, 32000)
        print('Wrote file (%s bytes)' % count)

    if args.chdir:
        os.chdir(args.chdir[0][0])
    result = subprocess.run(args.args, shell=False, check=False)
    exit(result.returncode)


if __name__ == '__main__':
    main()



# Copyright 2017 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import json

import paho.mqtt.client as mqtt

from zuul.connection import BaseConnection
from zuul.exceptions import ConfigurationError
from zuul.lib.logutil import get_annotated_logger


class MQTTConnection(BaseConnection):
    driver_name = 'mqtt'
    log = logging.getLogger("zuul.MQTTConnection")

    def __init__(self, driver, connection_name, connection_config):
        super(MQTTConnection, self).__init__(driver, connection_name,
                                             connection_config)
        self.client = mqtt.Client(
            client_id=self.connection_config.get('client_id'))
        if self.connection_config.get('user'):
            self.client.username_pw_set(
                self.connection_config.get('user'),
                self.connection_config.get('password'))
        ca_certs = self.connection_config.get('ca_certs')
        certfile = self.connection_config.get('certfile')
        keyfile = self.connection_config.get('keyfile')
        ciphers = self.connection_config.get('ciphers')
        if (ciphers or certfile or keyfile) and not ca_certs:
            raise ConfigurationError(
                "MQTT TLS configuration requires the ca_certs option")
        if ca_certs:
            if bool(certfile) != bool(keyfile):
                raise ConfigurationError(
                    "MQTT configuration keyfile and certfile "
                    "options must both be set.")
            self.client.tls_set(
                ca_certs,
                certfile=certfile,
                keyfile=keyfile,
                ciphers=ciphers)
        self.connected = False
        self.client.on_connect = self._on_connect
        self.client.on_disconnect = self._on_disconnect

    def _on_connect(self, client, userdata, flags, rc):
        self.connected = True

    def _on_disconnect(self, client, userdata, rc):
        self.connected = False

    def onLoad(self, zk_client, component_registry):
        self.log.debug("Starting MQTT Connection")

        # If the connection was not loaded by a scheduler, but by e.g.
        # zuul-web, we want to stop here.
        if not self.sched:
            return

        try:
            self.client.connect(
                self.connection_config.get('server', 'localhost'),
                port=int(self.connection_config.get('port', 1883)),
                keepalive=int(self.connection_config.get('keepalive', 60))
            )
            self.connected = True
        except Exception:
            self.log.exception("MQTT reporter (%s) couldn't connect" % self)
        self.client.loop_start()

    def onStop(self):
        if self.connected:
            self.log.debug("Stopping MQTT Connection")
            self.client.loop_stop()
            self.client.disconnect()
            self.connected = False

    def publish(self, topic, message, qos, zuul_event_id):
        log = get_annotated_logger(self.log, zuul_event_id)
        if not self.connected:
            log.warning("MQTT reporter (%s) is disabled", self)
            return
        try:
            self.client.publish(topic, payload=json.dumps(message), qos=qos)
        except Exception:
            log.exception(
                "Could not publish message to topic '%s' via mqtt", topic)



# Copyright 2021 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import threading

from zuul.zk.components import COMPONENT_REGISTRY
from zuul.zk.locks import SessionAwareLock, locked

import mmh3
from kazoo.protocol.states import KazooState
from kazoo.recipe.election import Election
from kazoo.exceptions import CancelledError


class SessionAwareElection(Election):
    def __init__(self, client, path, identifier=None):
        self._zuul_session_expired = False
        super().__init__(client, path, identifier)

    def run(self, func, *args, **kwargs):
        self._zuul_session_expired = False
        self.lock.client.add_listener(self._zuul_session_watcher)
        try:
            return super().run(func, *args, **kwargs)
        finally:
            self.lock.client.remove_listener(self._zuul_session_watcher)

    def _zuul_session_watcher(self, state):
        if state == KazooState.LOST:
            self._zuul_session_expired = True

    def is_still_valid(self):
        return not self._zuul_session_expired


def _default_filter(x):
    return x.state == x.RUNNING


class RendezvousElection:
    def __init__(self, client, path, component_type, this_component,
                 component_filter=None):
        self.log = logging.getLogger("zuul.RendezvousElection")
        self.client = client
        self.component_type = component_type
        self.component_filter = component_filter or _default_filter
        self.this_component = this_component
        self.client = client
        self.path = path
        self.lock = SessionAwareLock(self.client, self.path)
        # Whether we are the current winner of the rendezvous hash
        self.is_winner = None
        self.component_change_event = threading.Event()
        COMPONENT_REGISTRY.registry.registerCallback(self._onComponentChange)

    # Similar to the Election API
    def run(self, func, *args, **kw):
        # Allow the cancel method to stop this loop
        # Handle a restart
        self.running = True
        if self.is_winner is None:
            self._checkWinner()
        while self.running:
            if not self.is_winner:
                self.log.debug("Did not win election for %s", self.path)
                self.component_change_event.wait()
                self.component_change_event.clear()
                continue
            try:
                self.log.debug("Acquiring lock for %s", self.path)
                with locked(self.lock):
                    self.log.info("Won election for %s", self.path)
                    func(*args, **kw)
            except CancelledError:
                pass
            finally:
                self.log.info("Finished term for %s", self.path)

    # Similar to the Election API
    def cancel(self):
        self.running = False
        self.is_winner = None
        self.lock.cancel()
        self.component_change_event.set()

    # Similar to the Lock API
    def is_still_valid(self):
        if self.is_winner:
            return self.lock.is_still_valid()
        return False

    def _getScores(self):
        return {
            mmh3.hash(f"{c.hostname}-{self.path}", signed=False): c
            for c in COMPONENT_REGISTRY.registry.all(self.component_type)
            if self.component_filter(c)
        }

    def _getWinner(self):
        scores = self._getScores()
        winner = None
        if scores:
            sorted_scores = sorted(scores.keys())
            best_score = sorted_scores[0]
            winner = scores[best_score]
        return winner

    def _checkWinner(self):
        winner = self._getWinner()
        self.log.debug("Election winner for %s is %s", self.path, winner)
        self.is_winner = (
            winner and
            winner.hostname == self.this_component.hostname
        )

    def _onComponentChange(self):
        self._checkWinner()
        self.component_change_event.set()



# Copyright 2024-2025 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from collections import Counter
from unittest import mock

import testtools
from kazoo.exceptions import NoNodeError

from zuul import model
from zuul.launcher.client import LauncherClient
import zuul.launcher.server
import zuul.driver.static.staticendpoint
from tests.base import (
    ZuulTestCase,
    okay_tracebacks,
    iterate_timeout,
    simple_layout,
)
from tests.fake_nodescan import NodescanFixture


class TestStaticDriver(ZuulTestCase):
    config_file = 'zuul-connections-nodepool.conf'

    def _assertProviderNodeAttributes(self, pnode):
        self.assertEqual(pnode.connection_type, 'ssh')
        self.assertEqual(pnode.username, 'vinz')
        self.assertIsNotNone(pnode.interface_ip)
        self.assertIsNone(pnode.cloud)
        self.assertIsNone(pnode.region)
        if checks := self.test_config.driver.static.get('node_checks'):
            checks(self, pnode)

    def setUp(self):
        self.useFixture(NodescanFixture())
        super().setUp()

    def _getProvider(self):
        # Use the launcher provider so that we're using the same ttl
        # method caches.
        for provider in self.launcher.tenant_providers['tenant-one']:
            if provider.name == 'static-main':
                return provider

    def _waitForRegistration(self):
        for _ in iterate_timeout(60, "nodes to be ready"):
            nodes = self.launcher.api.nodes_cache.getItems()
            if (len(nodes) and
                all(n.state in (n.State.READY, n.State.SLOT_HOST)
                    for n in nodes)):
                break

    def _test_static_node_lifecycle(self, label):
        # Similar to the cloud test, but we don't deal with "instances"
        self._waitForRegistration()
        for _ in iterate_timeout(
                30, "scheduler and launcher to have the same layout"):
            if (self.scheds.first.sched.local_layout_state.get("tenant-one") ==
                self.launcher.local_layout_state.get("tenant-one")):
                break
        nodeset = model.NodeSet()
        nodeset.addNode(model.Node("node", label))

        ctx = self.createZKContext(None)
        request = self.requestNodes([n.label for n in nodeset.getNodes()])

        client = LauncherClient(self.zk_client, None)
        request = client.getRequest(request.uuid)

        self.assertEqual(request.state, model.NodesetRequest.State.FULFILLED)
        self.assertEqual(len(request.nodes), 1)

        client.acceptNodeset(request, nodeset)
        self.waitUntilSettled()

        with testtools.ExpectedException(NoNodeError):
            # Request should be gone
            request.refresh(ctx)

        for node in nodeset.getNodes():
            pnode = node._provider_node
            self.assertIsNotNone(pnode)
            self.assertTrue(pnode.hasLock())
            self._assertProviderNodeAttributes(pnode)

        client.useNodeset(nodeset)
        self.waitUntilSettled()

        for node in nodeset.getNodes():
            pnode = node._provider_node
            self.assertTrue(pnode.hasLock())
            self.assertEqual(pnode.state, pnode.State.IN_USE)

        client.returnNodeset(nodeset)
        self.waitUntilSettled()

        for node in nodeset.getNodes():
            pnode = node._provider_node
            self.assertFalse(pnode.hasLock())
            self.assertTrue(pnode.state in (pnode.State.USED,
                                            pnode.State.READY))

            for _ in iterate_timeout(60, "node to be marked ready"):
                pnode.refresh(ctx)
                if pnode.state == pnode.State.READY:
                    break

        # Make sure it can be reused
        request = self.requestNodes([label])
        self.assertEqual(request.state, model.NodesetRequest.State.FULFILLED)
        self.assertEqual(len(request.nodes), 1)

    @simple_layout('layouts/static/nodepool.yaml', enable_nodepool=True)
    def test_static_node_lifecycle(self):
        self._test_static_node_lifecycle('debian-normal')

    @simple_layout('layouts/static/nodepool.yaml', enable_nodepool=True)
    @okay_tracebacks('_checkNodescanRequest')
    def test_static_failure(self):
        # Test failure of a single static node
        ctx = self.createZKContext(None)

        def my_advance(*args, **kw):
            raise Exception("Test exception")

        with mock.patch.object(
            zuul.launcher.server.NodescanRequest, 'advance', my_advance
        ):
            request = self.requestNodes(["debian-normal"], timeout=None)

            for _ in iterate_timeout(60, "nodes to be marked failed"):
                nodes = self.launcher.api.nodes_cache.getItems()
                states = Counter(n.state for n in nodes)
                if states['failed'] == 1:
                    break

        # Node scanning happens normally now
        with mock.patch.object(
            zuul.driver.static.staticendpoint, 'DELETE_RECYCLE_TIME', 0
        ):
            for _ in iterate_timeout(60, "nodes to be marked ready"):
                request2 = self.launcher.api.getNodesetRequest(request.uuid)
                nodes = self.launcher.api.nodes_cache.getItems()
                states = Counter(n.state for n in nodes)
                if (states['ready'] == 1 and
                    not request2.is_locked):
                    break

        nodes = self.launcher.api.nodes_cache.getItems()
        self.assertEqual(1, len(nodes))
        main = nodes[0]
        # Get a copy so we're not modifying the launcher's
        main = model.ProviderNode.fromZK(ctx, path=main.getPath())
        self.assertEqual(main.State.READY, main.state)
        self.assertEqual(request2.uuid, main.request_id)

        request.delete(ctx)
        self.waitUntilSettled()

        for _ in iterate_timeout(60, "nodes to be unassigned"):
            nodes = self.launcher.api.nodes_cache.getItems()
            if not any(x.request_id for x in nodes):
                break

    @simple_layout('layouts/static/subnodes.yaml', enable_nodepool=True)
    def test_static_subnodes(self):
        self._waitForRegistration()
        ctx = self.createZKContext(None)
        request = self.requestNodes(['debian-normal'])
        self.assertEqual(request.State.FULFILLED, request.state)

        nodes = self.launcher.api.nodes_cache.getItems()
        self.assertEqual(3, len(nodes))
        # Get a list with the main node first and the subnodes last
        nodes.sort(key=lambda x: len(x.subnodes))
        nodes.reverse()
        main = nodes[0]
        sub1 = nodes[1]
        sub2 = nodes[2]
        # Get a copy so we're not modifying the launcher's
        main = model.ProviderNode.fromZK(ctx, path=main.getPath())
        sub1 = model.ProviderNode.fromZK(ctx, path=sub1.getPath())
        sub2 = model.ProviderNode.fromZK(ctx, path=sub2.getPath())
        self.assertIsNone(main.main_node_id)
        self.assertEqual(main.uuid, sub1.main_node_id)
        self.assertEqual(main.uuid, sub2.main_node_id)
        self.assertEqual(set([sub1.uuid, sub2.uuid]), set(main.subnodes))
        self.assertEqual([], sub1.subnodes)
        self.assertEqual([], sub2.subnodes)
        self.assertEqual(main.State.SLOT_HOST, main.state)
        self.assertEqual(sub1.State.READY, sub1.state)
        self.assertEqual(sub2.State.READY, sub2.state)

        with sub1.locked(ctx):
            sub1.refresh(ctx)
            with sub1.activeContext(ctx):
                sub1.assign(ctx, request_id="dne", tenant_name="test")
                # Modify the request so that the max zxid of the
                # request cache is later than the node assignment
                # (since the "dne" request does not exist).
                with request.locked(ctx), request.activeContext(ctx):
                    request.priority += 1
                sub1.setState(sub1.State.USED)

        for _ in iterate_timeout(60, "sub1 to be marked ready"):
            sub1.refresh(ctx)
            if sub1.state == sub1.State.READY:
                break

        # Main and sub2 should still exist
        main.refresh(ctx)
        sub2.refresh(ctx)

        with sub2.locked(ctx):
            sub2.refresh(ctx)
            with sub2.activeContext(ctx):
                sub2.unassign(ctx)
                sub2.assign(ctx, request_id="dne", tenant_name="test")
                # Modify the request so that the max zxid of the
                # request cache is later than the node assignment
                # (since the "dne" request does not exist).
                with request.locked(ctx), request.activeContext(ctx):
                    request.priority += 1
                sub2.setState(sub2.State.USED)

        for _ in iterate_timeout(60, "sub2 to be marked ready"):
            sub2.refresh(ctx)
            if sub2.state == sub2.State.READY:
                break

    @simple_layout('layouts/static/subnodes.yaml', enable_nodepool=True)
    @okay_tracebacks('_checkNodescanRequest')
    def test_static_subnodes_failure(self):
        # Test a node failure
        ctx = self.createZKContext(None)

        def my_advance(*args, **kw):
            raise Exception("Test exception")

        with mock.patch.object(
            zuul.launcher.server.NodescanRequest, 'advance', my_advance
        ):
            request = self.requestNodes(["debian-normal"], timeout=None)

            for _ in iterate_timeout(60, "nodes to be marked failed"):
                nodes = self.launcher.api.nodes_cache.getItems()
                states = Counter(n.state for n in nodes)
                if states['failed'] == 3:
                    break

        # Node scanning happens normally now
        with mock.patch.object(
            zuul.driver.static.staticendpoint, 'DELETE_RECYCLE_TIME', 0
        ):
            for _ in iterate_timeout(60, "nodes to be marked ready"):
                request2 = self.launcher.api.getNodesetRequest(request.uuid)
                nodes = self.launcher.api.nodes_cache.getItems()
                states = Counter(n.state for n in nodes)
                if (states['ready'] == 2 and states['slot-host'] == 1 and
                    not request2.is_locked):
                    break

        nodes = self.launcher.api.nodes_cache.getItems()
        self.assertEqual(3, len(nodes))
        # Get a list with the main node first and the subnode last
        nodes.sort(key=lambda x: len(x.subnodes))
        nodes.reverse()
        main = nodes[0]
        sub1 = nodes[1]
        sub2 = nodes[2]
        # Get a copy so we're not modifying the launcher's
        main = model.ProviderNode.fromZK(ctx, path=main.getPath())
        sub1 = model.ProviderNode.fromZK(ctx, path=sub1.getPath())
        sub2 = model.ProviderNode.fromZK(ctx, path=sub2.getPath())
        self.assertIsNone(main.main_node_id)
        self.assertEqual(main.uuid, sub1.main_node_id)
        self.assertEqual(main.uuid, sub2.main_node_id)
        self.assertEqual(2, len(main.subnodes))
        self.assertIn(sub1.uuid, main.subnodes)
        self.assertIn(sub2.uuid, main.subnodes)
        self.assertEqual([], sub1.subnodes)
        self.assertEqual([], sub2.subnodes)
        self.assertEqual(main.State.SLOT_HOST, main.state)
        self.assertEqual(sub1.State.READY, sub1.state)
        self.assertEqual(sub2.State.READY, sub2.state)
        if (sub1.request_id == request.uuid):
            self.assertIsNone(sub2.request_id)
        elif (sub2.request_id == request.uuid):
            self.assertIsNone(sub1.request_id)
        else:
            self.assertTrue(False, "one of the subnodes must be assigned")

        request.delete(ctx)
        self.waitUntilSettled()

        for _ in iterate_timeout(60, "nodes to be unassigned"):
            nodes = self.launcher.api.nodes_cache.getItems()
            if not any(x.request_id for x in nodes):
                break

    @simple_layout('layouts/static/subnodes.yaml', enable_nodepool=True)
    @okay_tracebacks('_checkNodescanRequest')
    def test_static_subnodes_reuse_failure(self):
        # Test that if we detect a failure in a subnode after it is
        # used once that we mark it and the main node as failed for
        # later deletion.
        ctx = self.createZKContext(None)
        request = self.requestNodes(['debian-normal'])
        self.assertEqual(request.State.FULFILLED, request.state)

        nodes = self.launcher.api.nodes_cache.getItems()
        self.assertEqual(3, len(nodes))
        # Get a list with the main node first and the subnode last
        nodes.sort(key=lambda x: len(x.subnodes))
        nodes.reverse()
        main = nodes[0]
        self.assertIsNone(main.main_node_id)
        subnodes = {x.uuid: x for x in nodes[1:]}
        sub1 = subnodes[main.subnodes[0]]
        sub2 = subnodes[main.subnodes[1]]
        # Get a copy so we're not modifying the launcher's
        main = model.ProviderNode.fromZK(ctx, path=main.getPath())
        sub1 = model.ProviderNode.fromZK(ctx, path=sub1.getPath())
        sub2 = model.ProviderNode.fromZK(ctx, path=sub2.getPath())
        self.assertEqual(main.uuid, sub1.main_node_id)
        self.assertEqual(main.uuid, sub2.main_node_id)
        self.assertEqual(set([sub1.uuid, sub2.uuid]), set(main.subnodes))
        self.assertEqual([], sub1.subnodes)
        self.assertEqual([], sub2.subnodes)
        self.assertEqual(main.State.SLOT_HOST, main.state)
        self.assertEqual(sub1.State.READY, sub1.state)
        self.assertEqual(sub2.State.READY, sub2.state)
        if (sub1.request_id == request.uuid):
            self.assertIsNone(sub2.request_id)
            sub_used = sub1
        elif (sub2.request_id == request.uuid):
            self.assertIsNone(sub1.request_id)
            sub_used = sub2
        else:
            self.assertTrue(False, "one of the subnodes must be assigned")

        def my_advance(*args, **kw):
            raise Exception("Test exception")

        with mock.patch.object(
            zuul.launcher.server.NodescanRequest, 'advance', my_advance
        ):
            with sub_used.locked(ctx):
                sub_used.refresh(ctx)
                with sub_used.activeContext(ctx):
                    sub_used.unassign(ctx)
                    sub_used.assign(ctx, request_id="dne", tenant_name="test")
                    # Modify the request so that the max zxid of the
                    # request cache is later than the node assignment
                    # (since the "dne" request does not exist).
                    with request.locked(ctx), request.activeContext(ctx):
                        request.priority += 1
                    sub_used.setState(sub_used.State.USED)

            for _ in iterate_timeout(10, "nodes to be marked failed"):
                nodes = self.launcher.api.nodes_cache.getItems()
                states = Counter(n.state for n in nodes)
                if states['failed'] == 3:
                    break

            # Delete the request
            request.delete(ctx)

        with mock.patch.object(
            zuul.driver.static.staticendpoint, 'DELETE_RECYCLE_TIME', 0
        ):
            for _ in iterate_timeout(10, "nodes to be marked ready"):
                nodes = self.launcher.api.nodes_cache.getItems()
                states = Counter(n.state for n in nodes)
                if (states['ready'] == 2 and states['slot-host'] == 1):
                    break

        # Make sure they are unassigned
        sub1.refresh(ctx)
        sub2.refresh(ctx)
        self.assertIsNone(sub1.request_id)
        self.assertIsNone(sub2.request_id)

    @simple_layout('layouts/static/nodepool.yaml', enable_nodepool=True)
    def test_static_deregistration(self):
        self._waitForRegistration()

        self.commitConfigUpdate(
            "org/common-config", 'layouts/static/nodepool-dereg.yaml')
        self.scheds.execute(lambda app: app.sched.reconfigure(app.config))
        self.waitUntilSettled()

        for _ in iterate_timeout(60, "nodes to be deregistered"):
            nodes = self.launcher.api.nodes_cache.getItems()
            if not len(nodes):
                break

    @simple_layout('layouts/static/nodepool.yaml', enable_nodepool=True)
    def test_static_deregistration_in_use(self):
        ctx = self.createZKContext(None)
        self._waitForRegistration()

        nodes = self.launcher.api.nodes_cache.getItems()
        main = nodes[0]
        # Get a copy so we don't share the lock with the launcher
        main = model.ProviderNode.fromZK(ctx, path=main.getPath())

        with main.locked(ctx):
            main.refresh(ctx)
            with main.activeContext(ctx):
                main.assign(ctx, request_id="dne", tenant_name="test")
                # Modify the request cache's max zxid to be later than
                # the node assignment (since the "dne" request does
                # not exist, and we have no requests in this test).
                self.launcher.api.requests_cache._max_zxid =\
                    main.min_request_zxid
                main.setState(main.State.USED)

            self.commitConfigUpdate(
                "org/common-config", 'layouts/static/nodepool-dereg.yaml')
            self.scheds.execute(lambda app: app.sched.reconfigure(app.config))
            self.waitUntilSettled()

        for _ in iterate_timeout(60, "nodes to be deregistered"):
            nodes = self.launcher.api.nodes_cache.getItems()
            if not len(nodes):
                break

    @simple_layout('layouts/static/subnodes.yaml', enable_nodepool=True)
    def test_static_subnodes_deregistration(self):
        self._waitForRegistration()

        self.commitConfigUpdate(
            "org/common-config", 'layouts/static/nodepool-dereg.yaml')
        self.scheds.execute(lambda app: app.sched.reconfigure(app.config))
        self.waitUntilSettled()

        for _ in iterate_timeout(60, "nodes to be deregistered"):
            nodes = self.launcher.api.nodes_cache.getItems()
            if not len(nodes):
                break

    @simple_layout('layouts/static/subnodes.yaml', enable_nodepool=True)
    def test_static_subnodes_deregistration_in_use(self):
        ctx = self.createZKContext(None)
        self._waitForRegistration()

        nodes = self.launcher.api.nodes_cache.getItems()
        # Get a list with the main node first and the subnodes last
        nodes.sort(key=lambda x: len(x.subnodes))
        nodes.reverse()
        sub1 = nodes[1]
        # Get a copy so we don't share the lock with the launcher
        sub1 = model.ProviderNode.fromZK(ctx, path=sub1.getPath())
        with sub1.locked(ctx):
            sub1.refresh(ctx)
            with sub1.activeContext(ctx):
                sub1.assign(ctx, request_id="dne", tenant_name="test")
                # Modify the request cache's max zxid to be later than
                # the node assignment (since the "dne" request does
                # not exist, and we have no requests in this test).
                self.launcher.api.requests_cache._max_zxid =\
                    sub1.min_request_zxid
                sub1.setState(sub1.State.USED)

            self.commitConfigUpdate(
                "org/common-config", 'layouts/static/nodepool-dereg.yaml')
            self.scheds.execute(lambda app: app.sched.reconfigure(app.config))
            self.waitUntilSettled()

        for _ in iterate_timeout(60, "nodes to be deregistered"):
            nodes = self.launcher.api.nodes_cache.getItems()
            if not len(nodes):
                break

    @simple_layout('layouts/static/nodepool.yaml', enable_nodepool=True)
    def test_static_node_aliases(self):
        self._test_static_node_lifecycle('special-hardware')



# Copyright 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from zuul.driver import Driver, ConnectionInterface, ReporterInterface
from zuul.driver.smtp import smtpconnection
from zuul.driver.smtp import smtpreporter


class SMTPDriver(Driver, ConnectionInterface, ReporterInterface):
    name = 'smtp'

    def getConnection(self, name, config):
        return smtpconnection.SMTPConnection(self, name, config)

    def getReporter(self, connection, pipeline, config=None,
                    parse_context=None):
        return smtpreporter.SMTPReporter(self, connection, config)

    def getReporterSchema(self):
        return smtpreporter.getSchema()



# Copyright 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import time
from uuid import uuid4

from opentelemetry import trace

from zuul.driver import Driver, TriggerInterface, ReporterInterface
from zuul.driver.zuul.zuulmodel import ZuulTriggerEvent
from zuul.driver.zuul import (
    zuulmodel,
    zuultrigger,
    zuulreporter,
)
from zuul.lib.logutil import get_annotated_logger
from zuul.model import Change

PARENT_CHANGE_ENQUEUED = 'parent-change-enqueued'
PROJECT_CHANGE_MERGED = 'project-change-merged'
IMAGE_BUILD = 'image-build'
IMAGE_VALIDATE = 'image-validate'
IMAGE_DELETE = 'image-delete'


class ZuulDriver(Driver, TriggerInterface, ReporterInterface):
    name = 'zuul'
    log = logging.getLogger("zuul.ZuulTrigger")
    tracer = trace.get_tracer("zuul")

    def __init__(self):
        self.parent_change_enqueued_events = {}
        self.project_change_merged_events = {}

    def registerScheduler(self, scheduler):
        self.sched = scheduler

    def reconfigure(self, tenant):
        for manager in tenant.layout.pipeline_managers.values():
            for ef in manager.pipeline.event_filters:
                if not isinstance(ef.trigger, zuultrigger.ZuulTrigger):
                    continue
                if PARENT_CHANGE_ENQUEUED in ef._types:
                    # parent-change-enqueued events need to be filtered by
                    # pipeline
                    for pipeline_name in ef._pipelines:
                        key = (tenant.name, pipeline_name)
                        self.parent_change_enqueued_events[key] = True
                elif PROJECT_CHANGE_MERGED in ef._types:
                    self.project_change_merged_events[tenant.name] = True

    def onChangeMerged(self, tenant, change, source):
        # Called each time zuul merges a change
        if self.project_change_merged_events.get(tenant.name):
            span = trace.get_current_span()
            link_attributes = {"rel": "ChangeMerged"}
            link = trace.Link(span.get_span_context(),
                              attributes=link_attributes)
            attributes = {"event_type": PROJECT_CHANGE_MERGED}
            with self.tracer.start_as_current_span(
                    "ZuulEvent", links=[link], attributes=attributes):
                try:
                    self._createProjectChangeMergedEvents(change, source)
                except Exception:
                    self.log.exception(
                        "Unable to create project-change-merged events for "
                        "%s" % (change,))

    def onChangeEnqueued(self, tenant, change, manager, event):
        log = get_annotated_logger(self.log, event)

        # Called each time a change is enqueued in a pipeline
        tenant_events = self.parent_change_enqueued_events.get(
            (tenant.name, manager.pipeline.name))
        log.debug("onChangeEnqueued %s", tenant_events)
        if tenant_events:
            span = trace.get_current_span()
            link_attributes = {"rel": "ChangeEnqueued"}
            link = trace.Link(span.get_span_context(),
                              attributes=link_attributes)
            attributes = {"event_type": PARENT_CHANGE_ENQUEUED}
            with self.tracer.start_as_current_span(
                    "ZuulEvent", links=[link], attributes=attributes):
                try:
                    self._createParentChangeEnqueuedEvents(
                        change, manager, tenant, event)
                except Exception:
                    log.exception(
                        "Unable to create parent-change-enqueued events for "
                        "%s in %s" % (change, manager.pipeline))

    def _createProjectChangeMergedEvents(self, change, source):
        changes = source.getProjectOpenChanges(
            change.project)
        for open_change in changes:
            self._createProjectChangeMergedEvent(open_change)

    def _createProjectChangeMergedEvent(self, change):
        event = ZuulTriggerEvent()
        event.type = PROJECT_CHANGE_MERGED
        event.trigger_name = self.name
        event.connection_name = "zuul"
        event.project_hostname = change.project.canonical_hostname
        event.project_name = change.project.name
        event.change_number = change.number
        event.branch = change.branch
        event.change_url = change.url
        event.patch_number = change.patchset
        event.ref = change.ref
        event.zuul_event_id = str(uuid4().hex)
        event.timestamp = time.time()
        self.sched.addTriggerEvent(self.name, event)

    def _createParentChangeEnqueuedEvents(self, change, manager, tenant,
                                          event):
        log = get_annotated_logger(self.log, event)

        log.debug("Checking for changes needing %s:" % change)
        if not isinstance(change, Change):
            log.debug("  %s does not support dependencies" % type(change))
            return

        # This is very inefficient, especially on systems with large
        # numbers of github installations.  This can be improved later
        # with persistent storage of dependency information.
        needed_by_changes = set(
            manager.resolveChangeReferences(change.getNeededByChanges()))
        for source in self.sched.connections.getSources():
            log.debug("  Checking source: %s",
                      source.connection.connection_name)
            new_changes = source.getChangesDependingOn(change, None, tenant)
            log.debug("  Source %s found %s changes",
                      source.connection.connection_name,
                      len(new_changes))
            needed_by_changes.update(new_changes)
        log.debug("  Following changes: %s", needed_by_changes)

        for needs in needed_by_changes:
            self._createParentChangeEnqueuedEvent(needs, manager)

    def _createParentChangeEnqueuedEvent(self, change, manager):
        event = ZuulTriggerEvent()
        event.type = PARENT_CHANGE_ENQUEUED
        event.connection_name = "zuul"
        event.trigger_name = self.name
        event.pipeline_name = manager.pipeline.name
        event.project_hostname = change.project.canonical_hostname
        event.project_name = change.project.name
        event.change_number = change.number
        event.branch = change.branch
        event.change_url = change.url
        event.patch_number = change.patchset
        event.ref = change.ref
        event.zuul_event_id = str(uuid4().hex)
        event.timestamp = time.time()
        self.sched.addTriggerEvent(self.name, event)

    def getImageBuildEvent(self, image_names, project_hostname, project_name,
                           project_branch):
        event = ZuulTriggerEvent()
        event.type = IMAGE_BUILD
        event.connection_name = "zuul"
        event.trigger_name = self.name
        event.image_names = image_names
        event.project_hostname = project_hostname
        event.project_name = project_name
        event.branch = project_branch
        event.ref = f'refs/heads/{project_branch}'
        event.zuul_event_id = str(uuid4().hex)
        event.timestamp = time.time()
        event.arrived_at_scheduler_timestamp = event.timestamp
        return event

    def getImageValidateEvent(self, image_names, project_hostname,
                              project_name, project_branch,
                              image_upload_uuid):
        event = ZuulTriggerEvent()
        event.type = IMAGE_VALIDATE
        event.connection_name = "zuul"
        event.trigger_name = self.name
        event.image_names = image_names
        event.image_upload_uuid = image_upload_uuid
        event.project_hostname = project_hostname
        event.project_name = project_name
        event.branch = project_branch
        event.ref = f'refs/heads/{project_branch}'
        event.zuul_event_id = str(uuid4().hex)
        event.timestamp = time.time()
        event.arrived_at_scheduler_timestamp = event.timestamp
        return event

    def getImageDeleteEvent(self, image_names, project_hostname, project_name,
                            project_branch, image_build_uuid):
        event = ZuulTriggerEvent()
        event.type = IMAGE_DELETE
        event.connection_name = "zuul"
        event.trigger_name = self.name
        event.image_names = image_names
        event.image_build_uuid = image_build_uuid
        event.project_hostname = project_hostname
        event.project_name = project_name
        event.branch = project_branch
        event.ref = f'refs/heads/{project_branch}'
        event.zuul_event_id = str(uuid4().hex)
        event.timestamp = time.time()
        event.arrived_at_scheduler_timestamp = event.timestamp
        return event

    def getTrigger(self, connection, config=None):
        return zuultrigger.ZuulTrigger(self, config)

    def getTriggerSchema(self):
        return zuultrigger.getSchema()

    def getTriggerEventClass(self):
        return zuulmodel.ZuulTriggerEvent

    def getReporter(self, connection, pipeline, config=None,
                    parse_context=None):
        return zuulreporter.ZuulReporter(self, connection, pipeline, config)

    def getReporterSchema(self):
        return zuulreporter.getSchema()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""add_uuid_indexes

Revision ID: e0eda5d09eae
Revises: c18b1277dfb5
Create Date: 2019-07-24 16:04:18.042209

"""

# revision identifiers, used by Alembic.
revision = 'e0eda5d09eae'
down_revision = 'c18b1277dfb5'
branch_labels = None
depends_on = None

from alembic import op

BUILDSET_TABLE = 'zuul_buildset'
BUILD_TABLE = 'zuul_build'


def upgrade(table_prefix=''):
    prefixed_buildset = table_prefix + BUILDSET_TABLE
    prefixed_build = table_prefix + BUILD_TABLE

    # To look up a build by uuid. buildset_id is included
    # so that it's a covering index and can satisfy the join back to buildset
    # without an additional lookup.
    op.create_index(
        table_prefix + 'uuid_buildset_id_idx', prefixed_build,
        ['uuid', 'buildset_id'])

    # To look up a buildset by uuid.
    op.create_index(table_prefix + 'uuid_idx', prefixed_buildset, ['uuid'])


def downgrade():
    raise Exception("Downgrades not supported")



# Copyright 2020 Red Hat Inc
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import argparse
import json
import sys
import datetime
import requests
from pathlib import Path


def usage(argv):
    two_weeks_ago = datetime.datetime.utcnow() - datetime.timedelta(days=14)
    parser = argparse.ArgumentParser(
        description="Look for unstrusted command in builds log")
    parser.add_argument(
        "--since", default=two_weeks_ago, help="Date in YYYY-MM-DD format")
    parser.add_argument("zuul_url", help="The url of a zuul-web service")
    args = parser.parse_args(argv)

    args.zuul_url = args.zuul_url.rstrip("/")
    if not args.zuul_url.endswith("/api"):
        args.zuul_url += "/api"
    if not isinstance(args.since, datetime.datetime):
        args.since = datetime.datetime.strptime(args.since, "%Y-%m-%d")
    return args


def get_tenants(zuul_url):
    """ Fetch list of tenant names """
    is_witelabel = requests.get(
        "%s/info" % zuul_url).json().get('tenant', None) is not None
    if is_witelabel:
        raise RuntimeError("Need multitenant api")
    return [
        tenant["name"]
        for tenant in requests.get("%s/tenants" % zuul_url).json()
    ]


def is_build_in_range(build, since):
    """ Check if a build is in range """
    try:
        build_date = datetime.datetime.strptime(
            build["start_time"], "%Y-%m-%dT%H:%M:%S")
        return build_date > since
    except TypeError:
        return False


def get_builds(zuul_builds_url, since):
    """ Fecth list of builds that are in range """
    builds = []
    pos = 0
    step = 50
    while not builds or is_build_in_range(builds[-1], since):
        url = "%s?skip=%d&limit=%d" % (zuul_builds_url, pos, step)
        print("Querying %s" % url)
        builds += requests.get(url).json()
        pos += step
    return builds


def filter_unique_builds(builds):
    """ Filter the list of build to keep only one per job name """
    jobs = dict()
    for build in builds:
        if build["job_name"] not in jobs:
            jobs[build["job_name"]] = build
    unique_builds = list(jobs.values())
    print("Found %d unique job builds" % len(unique_builds))
    return unique_builds


def download(source_url, local_filename):
    """ Download a file using streaming request """
    with requests.get(source_url, local_filename, stream=True) as r:
        r.raise_for_status()
        with open(local_filename, 'wb') as f:
            for chunk in r.iter_content(chunk_size=8192):
                f.write(chunk)


def download_build_job_output(zuul_build_url, local_path):
    """ Download the job-output.json of a build """
    build = requests.get(zuul_build_url).json()
    if not build.get("log_url"):
        return "No log url"
    try:
        download(build["log_url"] + "job-output.json", local_path)
    except Exception as e:
        return str(e)


def examine(path):
    """ Look for forbidden tasks in  a job-output.json file path """
    data = json.load(open(path))
    to_fix = False
    for playbook in data:
        if playbook['trusted']:
            continue
        for play in playbook['plays']:
            for task in play['tasks']:
                for hostname, host in task['hosts'].items():
                    if hostname != 'localhost':
                        continue
                    if host['action'] in ['command', 'shell']:
                        print("Found disallowed task:")
                        print("  Playbook: %s" % playbook['playbook'])
                        print("  Role: %s" % task.get('role', {}).get('name'))
                        print("  Task: %s" % task.get('task', {}).get('name'))
                        to_fix = True
    return to_fix


def main(argv):
    args = usage(argv)
    cache_dir = Path("/tmp/zuul-logs")
    if not cache_dir.exists():
        cache_dir.mkdir()

    to_fix = set()
    failed_to_examine = set()
    for tenant in get_tenants(args.zuul_url):
        zuul_tenant_url = args.zuul_url + "/tenant/" + tenant
        print("Looking for unique build in %s" % zuul_tenant_url)
        for build in filter_unique_builds(
                get_builds(zuul_tenant_url + "/builds", args.since)):
            if not build.get("uuid"):
                # Probably a SKIPPED build, no need to examine
                continue
            local_path = cache_dir / (build["uuid"] + ".json")
            build_url = zuul_tenant_url + "/build/" + build["uuid"]
            if not local_path.exists():
                err = download_build_job_output(build_url, str(local_path))
                if err:
                    failed_to_examine.add((build_url, err))
                    continue
            try:
                if not examine(str(local_path)):
                    print("%s: ok" % build_url)
                else:
                    to_fix.add(build_url)
            except Exception as e:
                failed_to_examine.add((build_url, str(e)))

    if failed_to_examine:
        print("The following builds could not be examined:")
        for build_url, err in failed_to_examine:
            print("%s: %s" % (build_url, err))
        if not to_fix:
            exit(1)

    if to_fix:
        print("The following builds are using localhost command:")
        for build in to_fix:
            print(build.replace("/api/", "/t/"))
        exit(1)


if __name__ == "__main__":
    main(sys.argv[1:])



# Copyright 2021 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from zuul.lib import yamlutil
from tests.base import BaseTestCase

import testtools


class TestYamlDumper(BaseTestCase):
    def test_load_normal_data(self):
        expected = {'foo': 'bar'}
        data = 'foo: bar\n'
        out = yamlutil.safe_load(data)
        self.assertEqual(out, expected)

        out = yamlutil.encrypted_load(data)
        self.assertEqual(out, expected)

    def test_load_encrypted_data(self):
        expected = {'foo': yamlutil.EncryptedPKCS1_OAEP('YmFy')}
        self.assertEqual(expected['foo'].ciphertext, b'bar')
        data = "foo: !encrypted/pkcs1-oaep YmFy\n"

        out = yamlutil.encrypted_load(data)
        self.assertEqual(out, expected)

        with testtools.ExpectedException(
                yamlutil.yaml.constructor.ConstructorError):
            out = yamlutil.safe_load(data)

    def test_dump_normal_data(self):
        data = {'foo': 'bar'}
        expected = 'foo: bar\n'
        out = yamlutil.safe_dump(data, default_flow_style=False)
        self.assertEqual(out, expected)

        out = yamlutil.encrypted_dump(data, default_flow_style=False)
        self.assertEqual(out, expected)

    def test_dump_encrypted_data(self):
        data = {'foo': yamlutil.EncryptedPKCS1_OAEP('YmFy')}
        self.assertEqual(data['foo'].ciphertext, b'bar')
        expected = "foo: !encrypted/pkcs1-oaep YmFy\n"

        out = yamlutil.encrypted_dump(data, default_flow_style=False)
        self.assertEqual(out, expected)

        with testtools.ExpectedException(
                yamlutil.yaml.representer.RepresenterError):
            out = yamlutil.safe_dump(data, default_flow_style=False)

    def test_ansible_dumper(self):
        data = {'foo': 'bar'}
        data = yamlutil.mark_strings_unsafe(data)
        expected = "foo: !unsafe bar\n"
        yaml_out = yamlutil.ansible_unsafe_dump(data, default_flow_style=False)
        # Assert the serialized string looks good
        self.assertEqual(yaml_out, expected)
        # Check the round trip
        data_in = yamlutil.ansible_unsafe_load(yaml_out)
        self.assertEqual(data, data_in)

        data = {'foo': {'bar': 'baz'}, 'list': ['bar', 1, 3.0, True, None]}
        data = yamlutil.mark_strings_unsafe(data)
        expected = """\
foo:
  bar: !unsafe baz
list:
- !unsafe bar
- 1
- 3.0
- true
- null
"""
        yaml_out = yamlutil.ansible_unsafe_dump(data, default_flow_style=False)
        # Assert the serialized string looks good
        self.assertEqual(yaml_out, expected)
        data_in = yamlutil.ansible_unsafe_load(yaml_out)
        # Check the round trip
        self.assertEqual(data, data_in)

    def test_ansible_dumper_with_aliases(self):
        foo = {'bar': 'baz'}
        data = {'foo1': foo, 'foo2': foo}
        expected = """\
foo1: &id001
  bar: baz
foo2: *id001
"""
        yaml_out = yamlutil.ansible_unsafe_dump(data, default_flow_style=False)
        self.assertEqual(yaml_out, expected)

    def test_ansible_dumper_ignore_aliases(self):
        foo = {'bar': 'baz'}
        data = {'foo1': foo, 'foo2': foo}
        expected = """\
foo1:
  bar: baz
foo2:
  bar: baz
"""
        yaml_out = yamlutil.ansible_unsafe_dump(
            data,
            ignore_aliases=True,
            default_flow_style=False)
        self.assertEqual(yaml_out, expected)



"""Allow score to be null

Revision ID: 1dd914d4a482
Revises: 4d3ebd7f06b9
Create Date: 2017-03-28 08:09:32.908643

"""

# revision identifiers, used by Alembic.
revision = '1dd914d4a482'
down_revision = '4d3ebd7f06b9'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade(table_prefix=''):
    op.alter_column(table_prefix + 'zuul_buildset', 'score', nullable=True,
                    existing_type=sa.Integer)


def downgrade():
    raise Exception("Downgrades not supported")



# Copyright 2016 Red Hat, Inc.
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.

import importlib
import os
import sys

import ansible.plugins.action


def _full_path(path):
    return os.path.realpath(os.path.abspath(os.path.expanduser(path)))


def _fail_dict(path, prefix='Accessing files from'):
    return dict(
        failed=True,
        path=path,
        msg="{prefix} outside the working dir {curdir} is prohibited".format(
            prefix=prefix,
            curdir=os.path.abspath(os.path.curdir)))


def _import_ansible_action_plugin(name):
    # Ansible forces the import of our action plugins
    # (zuul.ansible.action.foo) as ansible.plugins.action.foo, which
    # is the import path of the ansible implementation.  Our
    # implementations need to subclass that, but if we try to import
    # it with that name, we will get our own module.  This bypasses
    # Python's module namespace to load the actual ansible modules.
    # We need to give it a name, however.  If we load it with its
    # actual name, we will end up overwriting our module in Python's
    # namespace, causing infinite recursion.  So we supply an
    # otherwise unused name for the module:
    # zuul.ansible.protected.action.foo.
    #
    # From https://discuss.python.org/t/how-do-i-migrate-from-imp/27885/3
    # for converting imp module loads to python3.12 compatible code.
    spec = importlib.machinery.PathFinder.find_spec(
        name, ansible.plugins.action.__path__)
    if not spec:
        raise Exception("Could not find %s at module path %s" %
                        (name, ansible.plugins.action.__path__))
    mod = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(mod)
    sys.modules['zuul.ansible.protected.action.' + name] = mod
    return mod


def _sanitize_filename(name):
    return ''.join(c for c in name if c.isalnum())


# Ansible assigns a unique id to every task (Task._uuid).  However, if
# a role is included more than once, the task object is re-used.  In
# order to provide unique log ids for the Zuul command log streaming
# system, this global dictionary is used to map keys that are derived
# from tasks (task._uuid concatenated with the host name) to a counter
# which is incremented each time the task+host combination is
# encountered.  Ansible will not run more than one task on a host
# simultaneously, so this should be sufficiently unique to avoid
# collisions.
#
# We use a global dictionary defined here so that zuul_stream can
# write to it and zuul.ansible.command modules can read it.  Note that
# the command module operates after a fork and therefore it should be
# treated as read-only there.
ZUUL_LOG_ID_MAP = {}
